"""unittests for schema2dot"""

import os
import os.path as osp

from logilab.common.testlib import TestCase, unittest_main
from logilab.common.compat import set

from yams.reader import SchemaLoader
from yams import schema2dot

DATADIR = osp.abspath(osp.join(osp.dirname(__file__), 'data'))

schema = SchemaLoader().load([DATADIR])

DOT_SOURCE = r'''digraph "toto" {
rankdir=BT
ratio=compress
charset="utf-8"
splines=true
overlap=false
sep=0.2
"Person" [shape="record", fontname="Courier", style="filled", label="{Person|nom\lprenom\lsexe\lpromo\ltitre\ladel\lass\lweb\ltel\lfax\ldatenaiss\ltest\lsalary\l}"];
"Societe" [shape="record", fontname="Courier", style="filled", label="{Societe|nom\lweb\ltel\lfax\lrncs\lad1\lad2\lad3\lcp\lville\l}"];
"Person" -> "Societe" [taillabel="0..n", style="filled", arrowhead="normal", color="#aa0000", label="travaille", headlabel="0..n", arrowtail="none", decorate="false", fontcolor="#aa0000"];
}
'''


class DotTC(TestCase):

    def test_schema2dot(self):
        """tests dot conversion without attributes information"""
        wanted_entities = set(('Person', 'Societe'))
        skipped_entities = set(ent for ent in schema.entities()
                               if ent.type not in wanted_entities)
        schema2dot.schema2dot(schema, 'toto.dot', skiptypes=skipped_entities)
        generated = open('toto.dot').read()
        os.remove('toto.dot')
        self.assertTextEquals(DOT_SOURCE, generated)

if __name__ == '__main__':
    unittest_main()



import yams
yams.use_py_datetime()
from yams.constraints import *
# after import *
from datetime import datetime, date, timedelta

from logilab.common.testlib import TestCase, unittest_main, mock_object

class ConstraintTC(TestCase):

    def test_interval_serialization_integers(self):
        cstr = IntervalBoundConstraint(12, 13)
        self.assertEquals(cstr.serialize(), '12;13')
        cstr = IntervalBoundConstraint(maxvalue=13)
        self.assertEquals(cstr.serialize(), 'None;13')
        cstr = IntervalBoundConstraint(minvalue=13)
        self.assertEquals(cstr.serialize(), '13;None')
        self.assertRaises(AssertionError, IntervalBoundConstraint)

    def test_interval_serialization_floats(self):
        cstr = IntervalBoundConstraint(12.13, 13.14)
        self.assertEquals(cstr.serialize(), '12.13;13.14')


    def test_interval_deserialization_integers(self):
        cstr = IntervalBoundConstraint.deserialize('12;13')
        self.assertEquals(cstr.minvalue, 12)
        self.assertEquals(cstr.maxvalue, 13)
        cstr = IntervalBoundConstraint.deserialize('None;13')
        self.assertEquals(cstr.minvalue, None)
        self.assertEquals(cstr.maxvalue, 13)
        cstr = IntervalBoundConstraint.deserialize('12;None')
        self.assertEquals(cstr.minvalue, 12)
        self.assertEquals(cstr.maxvalue, None)

    def test_interval_deserialization_floats(self):
        cstr = IntervalBoundConstraint.deserialize('12.13;13.14')
        self.assertEquals(cstr.minvalue, 12.13)
        self.assertEquals(cstr.maxvalue, 13.14)


    def test_regexp_serialization(self):
        cstr = RegexpConstraint('[a-z]+,[A-Z]+', 12)
        self.assertEquals(cstr.serialize(), '[a-z]+,[A-Z]+,12')

    def test_regexp_deserialization(self):
        cstr = RegexpConstraint.deserialize('[a-z]+,[A-Z]+,12')
        self.assertEquals(cstr.regexp, '[a-z]+,[A-Z]+')
        self.assertEquals(cstr.flags, 12)

    def test_constraints_with_attribute(self):
        cstr = IntervalBoundConstraint(NOW(), Attribute('hop'))
        cstr2 = IntervalBoundConstraint.deserialize(cstr.serialize())
        self.assertEquals(cstr2.minvalue.offset, None)
        self.assertEquals(cstr2.maxvalue.attr, 'hop')
        self.failUnless(cstr2.check(mock_object(hop=datetime.now()+timedelta(hours=1)),
                                    'hip', datetime.now() + timedelta(seconds=2)))
        # fail, value < minvalue
        self.failIf(cstr2.check(mock_object(hop=datetime.now()+timedelta(hours=1)),
                                'hip', datetime.now() - timedelta(hours=2)))
        # fail, value > maxvalue
        self.failIf(cstr2.check(mock_object(hop=datetime.now()+timedelta(hours=1)),
                                'hip', datetime.now() + timedelta(hours=2)))


    def test_constraints_with_date(self):
        cstr = IntervalBoundConstraint(TODAY(timedelta(1)),
                                       TODAY(timedelta(3)))
        cstr2 = IntervalBoundConstraint.deserialize(cstr.serialize())
        self.assertEquals(cstr2.minvalue.offset, timedelta(1))
        self.assertEquals(cstr2.maxvalue.offset, timedelta(3))
        self.failUnless(cstr2.check(None, 'hip', date.today() + timedelta(2)))
        # fail, value < minvalue
        self.failIf(cstr2.check(None, 'hip', date.today()))
        # fail, value > maxvalue
        self.failIf(cstr2.check(None, 'hip', date.today() + timedelta(4)))

if __name__ == '__main__':
    unittest_main()





# -*- coding: iso-8859-1 -*-
"""unit tests for module yams.schema classes"""

from logilab.common.testlib import TestCase, unittest_main

from tempfile import mktemp

from yams.buildobjs import register_base_types, EntityType, RelationType, RelationDefinition
from yams.schema import *
from yams.constraints import *
from yams.reader import SchemaLoader


# build a dummy schema ########################################################


class BaseSchemaTC(TestCase):
    def setUp(self):
        global schema, enote, eaffaire, eperson, esociete, estring, eint
        global rconcerne, rnom
        schema = Schema('Test Schema')
        register_base_types(schema)
        enote = schema.add_entity_type(EntityType('Note'))
        eaffaire = schema.add_entity_type(EntityType('Affaire'))
        eperson = schema.add_entity_type(EntityType('Person'))
        esociete = schema.add_entity_type(EntityType('Societe'))


        RELS = (
            # attribute relations
            ('Note date Datetime'),
            ('Note type String'),
            ('Affaire sujet String'),
            ('Affaire ref String'),
            ('Affaire starton Time'),
            ('Person nom String'),
            ('Person prenom String'),
            ('Person sexe Float'),
            ('Person tel Int'),
            ('Person fax Int'),
            ('Person datenaiss Date'),
            ('Person TEST Boolean'),
            ('Person promo String'),
            ('Person promo_enlarged String'),
            ('Person promo_encoding String'),
            # real relations
            ('Person  travaille Societe'),
            ('Person  evaluee   Note'),
            ('Societe evaluee   Note'),
            ('Person  concerne  Affaire'),
            ('Person  concerne  Societe'),
            ('Affaire concerne  Societe'),
            )
        for i, rel in enumerate(RELS):
            _from, _type, _to = rel.split()
            try:
                schema.rschema(_type)
            except KeyError:
                schema.add_relation_type(RelationType(_type))
            schema.add_relation_def(RelationDefinition(_from, _type, _to, order=i))
        schema.rschema('nom').set_rproperty('Person', 'String', 'cardinality','11') # not null

        enote.set_rproperty('type', 'constraints',
                             [StaticVocabularyConstraint((u'bon', u'pasbon',
                                                                  u'bof', u'peux mieux faire'))])
        enote.set_rproperty('date', 'cardinality', '11')

        eaffaire.set_rproperty('sujet', 'constraints', [SizeConstraint(128)])
        eaffaire.set_rproperty('ref', 'constraints', [SizeConstraint(12), RegexpConstraint(r'[A-Z]+\d+')])
        eperson.set_rproperty('nom', 'constraints', [SizeConstraint(20, 10)])
        eperson.set_rproperty('prenom', 'constraints', [SizeConstraint(64)])
        eperson.set_rproperty('tel', 'constraints', [IntervalBoundConstraint(maxvalue=999999)])
        eperson.set_rproperty('fax', 'constraints', [IntervalBoundConstraint(minvalue=12, maxvalue=999999)])
        eperson.set_rproperty('promo', 'constraints', [StaticVocabularyConstraint( (u'bon', u'pasbon'))])

        estring = schema.eschema('String')
        eint = schema.eschema('Int')
        rconcerne = schema.rschema('concerne')
        rnom = schema.rschema('nom')

    def assertRaisesMsg(self, ex_class, msg, func, *args, **kwargs):
        self.assertRaises(ex_class, func, *args, **kwargs)
        try:
            func(*args, **kwargs)
        except Exception, ex:
            self.assertEquals(str(ex), msg)

# test data ###################################################################

BAD_RELS = ( ('Truc badrelation1 Affaire'),
             ('Affaire badrelation2 Machin'),
             )

ATTRIBUTE_BAD_VALUES = (
    ('Person', [('nom', 1), ('nom', u'tropcour'),
                ('nom', u'>10 mais  supérieur à < 20 , c\'est long'),
                ('sexe', u'F'), ('sexe', u'MorF'), ('sexe', 'F'),
                ('promo', 'bon'), ('promo', 'uyou'),
                ('promo', u' pas bon du tout'),
                ('tel', 'notastring'),
                ('tel', 1000000),
                ('fax', 11),
                ('TEST', 'notaboolean'), #('TEST', 0), ('TEST', 1)]), #should we accept this ?
                ('TEST', 'true'), ('TEST', 'false')]),
## the date and time are not checked for now
##    ('Person', [('nom', u' >10 mais < 20 '),
##               ('datenaiss', '979-06-12')]),
##    ('Note', [('date', '2229-01-31 minuit')]),
##    ('Affaire', [('starton', 'midi')]),

    ('Note', [('type', ['bof', 'peux mieux faire']),
              ('type', 'bof, je suis pas unicode, alors...'),
              ('date', None),
              ]),
    ('Affaire', [('ref', 'ginco01'), ('ref', 'GINCO'),],
    ),
    )

ATTRIBUTE_GOOD_VALUES = (
    ('Person', [('nom', u'>10 mais < 20 '), ('sexe', 0.5),
                ('promo', u'bon'),
                ('datenaiss', '1977-06-07'),
                ('tel', 83433), ('fax', None), ('fax', 12),
                ('TEST', True), ('TEST', False)]),
    ('Note', [('date', '2229-01-31 00:00')]),
    ('Affaire', [('starton', '00:00'),
                 ('ref', u'GINCO01')]),
    )

RELATIONS_BAD_VALUES = {
    'travaille': [('Person', 'Affaire'), ('Affaire', 'Societe'),
                  ('Affaire', 'Societe'), ('Societe', 'Person')]
    }
RELATIONS_GOOD_VALUES = {
    'travaille': [('Person', 'Societe')],
    'concerne': [('Person', 'Affaire'), ('Affaire', 'Societe')]
    }


# test suite ##################################################################

class EntitySchemaTC(BaseSchemaTC):

    def test_base(self):
        self.assert_(repr(eperson))

    def test_cmp(self):
        self.failUnless(eperson == 'Person')
        self.failUnless('Person' == eperson)
        self.failUnless(eperson != 'Note')
        self.failUnless('Note' != eperson)
        self.failIf(enote == eperson)
        self.failIf(eperson == enote)
        self.failUnless(enote != eperson)
        self.failUnless(eperson != enote)
        l = [eperson, enote, eaffaire, esociete]
        l.sort()
        self.assertListEquals(l, [eaffaire, enote, eperson, esociete])
        self.assertListEquals(l, ['Affaire', 'Note', 'Person', 'Societe'])

    def test_hash(self):
        from copy import copy
        d = {}
        d[eperson] = 'p'
        d[enote] = 'n'
        self.failUnlessEqual(d[copy(eperson)], 'p')
        self.failUnlessEqual(d[copy(enote)], 'n')
        self.failUnlessEqual(d['Person'], 'p')
        self.failUnlessEqual(d['Note'], 'n')
        d = {}
        d['Person'] = eperson
        d['Note'] = enote
        self.failUnlessEqual(copy(eperson), 'Person')
        self.failUnlessEqual(d[copy(eperson)], 'Person')
        self.failUnlessEqual(d[copy(enote)], 'Note')

    def test_deepcopy_with_regexp_constraints(self):
        from copy import deepcopy
        eaffaire.set_rproperty('ref', 'constraints', [RegexpConstraint(r'[A-Z]+\d+')])
        rgx_cstr, = eaffaire.constraints('ref')
        eaffaire2 = deepcopy(schema).eschema('Affaire')
        rgx_cstr2, = eaffaire2.constraints('ref')
        self.assertEquals(rgx_cstr2.regexp, rgx_cstr.regexp)
        self.assertEquals(rgx_cstr2.flags, rgx_cstr.flags)
        self.assertEquals(rgx_cstr2._rgx, rgx_cstr._rgx)

    def test_deepcopy(self):
        from copy import deepcopy
        global schema
        schema = deepcopy(schema)
        self.failIf(eperson is schema['Person'])
        self.failUnlessEqual(eperson, schema['Person'])
        self.failUnlessEqual('Person', schema['Person'])
        self.failUnlessEqual(eperson.subject_relations(), schema['Person'].subject_relations())
        self.failUnlessEqual(eperson.object_relations(), schema['Person'].object_relations())

    def test_deepcopy_specialization(self):
        schema2 = deepcopy(SchemaLoader().load([self.datadir], 'Test'))
        edivision = schema2.eschema('Division')
        self.assertEquals(edivision.specializes(), 'Company')
        self.assertEquals(edivision.specialized_by(), ['Subdivision'])

    def test_is_final(self):
        self.assertEquals(eperson.is_final(), False)
        self.assertEquals(enote.is_final(), False)
        self.assertEquals(estring.is_final(), True)
        self.assertEquals(eint.is_final(), True)
        self.assertEquals(eperson.subject_relation('nom').is_final(), True)
        #self.assertEquals(eperson.is_final('concerne'), False)
        self.assertEquals(eperson.subject_relation('concerne').is_final(), False)

    def test_is_metadata(self):
        self.assertEquals(eperson.is_metadata('promo'), None)
        self.assertEquals(eperson.is_metadata('promo_enlarged'), None)
        self.assertEquals(eperson.is_metadata('promo_encoding'), ('promo', 'encoding'))
        self.assertEquals([(k.type, v)  for k, v in eperson.meta_attributes().items()],
                          [('promo_encoding', ('encoding', 'promo'))])

    def test_defaults(self):
        self.assertEquals(list(eperson.defaults()), [])
        self.assertRaises(AssertionError,
                          estring.defaults().next)

    def test_vocabulary(self):
        self.assertEquals(eperson.vocabulary('promo'), ('bon', 'pasbon'))
        self.assertRaises(KeyError,
                          eperson.vocabulary, 'what?')
        self.assertRaises(AssertionError,
                          eperson.vocabulary, 'nom')

    def test_indexable_attributes(self):
        eperson.set_rproperty('nom', 'fulltextindexed', True)
        eperson.set_rproperty('prenom', 'fulltextindexed', True)
        self.assertEquals(list(eperson.indexable_attributes()), ['nom', 'prenom'])


    def test_goodValues_relation_default(self):
        """check good values of entity does not raise an exception"""
        eperson.set_rproperty('nom', 'default', 'No name')
        self.assertEquals(eperson.default('nom'), 'No name')

    def test_subject_relations(self):
        """check subject relations a returned in the same order as in the
        schema definition"""
        rels = eperson.ordered_relations()
        expected = ['nom', 'prenom', 'sexe', 'tel', 'fax', 'datenaiss',
                    'TEST', 'promo', 'promo_enlarged', 'promo_encoding', 'travaille',
                    'evaluee', 'concerne']
        self.assertEquals([r.type for r in rels], expected)

    def test_object_relations(self):
        """check object relations a returned in the same order as in the
        schema definition"""
        rels = eaffaire.object_relations()
        expected = ['concerne']
        self.assertEquals(rels, expected)
        rels = [schem.type for schem in eaffaire.object_relations()]
        self.assertEquals(rels, expected)
        self.assertEquals(eaffaire.object_relation('concerne').type,
                         'concerne')

    def test_destination_type(self):
        """check subject relations a returned in the same order as in the
        schema definition"""
        self.assertEquals(eperson.destination('nom'), 'String')
        self.assertRaises(AssertionError, eperson.destination, 'travaille')


class RelationSchemaTC(BaseSchemaTC):

    def test_cmp(self):
        self.failUnless(rconcerne == 'concerne')
        self.failUnless('concerne' == rconcerne)
        self.failUnless(rconcerne != 'nom')
        self.failUnless('nom' != rconcerne)
        self.failIf(rnom == rconcerne)
        self.failIf(rconcerne == rnom)
        self.failUnless(rnom != rconcerne)
        self.failUnless(rconcerne != rnom)

    def test_hash(self):
        from copy import copy
        d = {}
        d[rconcerne] = 'p'
        d[rnom] = 'n'
        self.failUnlessEqual(d[copy(rconcerne)], 'p')
        self.failUnlessEqual(d[copy(rnom)], 'n')
        self.failUnlessEqual(d['concerne'], 'p')
        self.failUnlessEqual(d['nom'], 'n')


    def test_base(self):
        self.assert_(repr(rnom))

    def test_star_types(self):
        types = sorted(rconcerne.subjects())
        self.assertEquals(types, ['Affaire', 'Person'])
        types = sorted(rconcerne.objects())
        self.assertEquals(types, ['Affaire', 'Societe'])

    def test_raise_update(self):
        self.assertRaisesMsg(BadSchemaDefinition,
                             'type String can\'t be used as subject in a relation',
                             rconcerne.update, estring, enote, {})
##         self.assertRaisesMsg(BadSchemaDefinition,
##                              "can't have a final relation pointing to multiple entity types (nom: ['String', 'Int'])" ,
##                              rnom.update, enote, eint)
        self.assertRaisesMsg(BadSchemaDefinition, 'ambiguous relation nom: String is final but not Affaire',
                             rnom.update, enote, eaffaire, {})
        self.assertRaises(BadSchemaDefinition,
                          rconcerne.update, enote, estring, {})

    def test_association_types(self):
        expected = [ ('Affaire', ['Societe']),
                     ('Person', ['Affaire', 'Societe']) ]
        assoc_types = rconcerne.associations()
        assoc_types.sort()
        self.assertEquals(assoc_types, expected)
        assoc_types = []
        for _from, _to in rconcerne.associations():
            assoc_types.append( (_from, _to))
            #assoc_types.append( (_from.type, [s.type for s in _to]) )
        assoc_types.sort()
        self.assertEquals(assoc_types, expected)

#     def test_reverse_association_types(self):
#         expected = [ ('Affaire', ['Person']),
#                      ('Societe', ['Person', 'Affaire'])]
#         assoc_types = rconcerne.reverse_association_types()
#         assoc_types.sort()
#         self.assertEquals(assoc_types, expected)
#         assoc_types = []
#         for _from, _to in rconcerne.reverse_association_types(True):
#             assoc_types.append( (_from.type, [s.type for s in _to]) )
#         assoc_types.sort()
#         self.assertEquals(assoc_types, expected)


class SchemaTC(BaseSchemaTC):

    def test_schema_base(self):
        """test base schema methods
        """
        all_types = ['Affaire', 'Boolean', 'Bytes', 'Date', 'Datetime',
                     'Decimal',
                     'Float', 'Int', 'Interval', 'Note', 'Password',
                     'Person', 'Societe', 'String', 'Time']
        types = schema.entities()
        types.sort()
        self.assertEquals(types, all_types)
        self.assertEquals(schema.has_entity('Affaire'), True)
        self.assertEquals(schema.has_entity('Aaire'), False)

    def test_raise_add_entity_type(self):
        self.assertRaisesMsg(BadSchemaDefinition, "entity type Person is already defined" ,
                             schema.add_entity_type, EntityType('Person'))

    def test_raise_relation_def(self):
        self.assertRaisesMsg(BadSchemaDefinition, "using unknown type 'Afire' in relation evaluee"  ,
                             schema.add_relation_def, RelationDefinition('Afire', 'evaluee', 'Note'))
## XXX what is this ?
##        self.assertRaisesMsg(BadSchemaDefinition, 'the "symetric" property should appear on every definition of relation evaluee' ,
##                             schema.add_relation_def, RelationDefinition('Affaire', 'evaluee', 'Note', symetric=True))

    def test_schema_relations(self):
        all_relations = ['TEST', 'concerne', 'travaille', 'evaluee',
                         'date', 'type', 'sujet', 'ref', 'nom', 'prenom',
                         'starton', 'sexe', 'promo', 'promo_enlarged',
                         'promo_encoding', 'tel', 'fax', 'datenaiss']
        all_relations.sort()
        relations = schema.relations()
        relations.sort()
        self.assertEquals(relations, all_relations)

        self.assertEquals(len(eperson.rproperty('nom', 'constraints')), 1)
        self.assertEquals(len(eperson.rproperty('prenom', 'constraints')), 1)

    def test_schema_check_relations(self):
        """test behaviour with some incorrect relations"""
        for rel in BAD_RELS:
            _from, _type, _to = rel.split()
            self.assertRaises(BadSchemaDefinition,
                              schema.add_relation_def, RelationDefinition(_from, _type, _to))
        # check we can't extend a final relation
        self.assertRaises(BadSchemaDefinition,
                          schema.add_relation_def, RelationDefinition('Person', 'nom', 'affaire'))

    def test_entities_goodValues_check(self):
        """check good values of entity does not raise an exception"""
        for etype, val_list in ATTRIBUTE_GOOD_VALUES:
            eschema = schema.eschema(etype)
            eschema.check(dict(val_list))

    def test_entities_badValues_check(self):
        """check bad values of entity raises ValidationError exception"""
        for etype, val_list in ATTRIBUTE_BAD_VALUES:
            eschema = schema.eschema(etype)
            # check attribute values one each time...
            for item in val_list:
                self.assertRaises(ValidationError, eschema.check, dict([item]))

    def test_pickle(self):
        """schema should be pickeable"""
        import pickle
        picklefile = mktemp()
        picklestream = open(picklefile, 'w')
        schema.__hashmode__ = 'pickle'
        pickle.dump(schema, picklestream)
        picklestream.close()
        pschema = pickle.load(open(picklefile))
        schema.__hashmode__ = None
        self.assertEquals(pschema.__hashmode__, None)
        self.failIf(eperson is pschema['Person'])
        self.failUnlessEqual(eperson, pschema['Person'])
        self.failUnlessEqual('Person', pschema['Person'])
        self.failUnlessEqual(eperson.ordered_relations(), pschema['Person'].ordered_relations())
        self.failUnlessEqual(eperson.object_relations(), pschema['Person'].object_relations())


    def test_rename_entity_type(self):
        affaire = schema.eschema('Affaire')
        orig_rprops = affaire.rproperties('concerne')
        schema.rename_entity_type('Affaire', 'Workcase')
        self.assertUnorderedIterableEquals(schema._entities.keys(),
                             ['Boolean', 'Bytes', 'Date', 'Datetime', 'Float',
                              'Decimal',
                              'Int', 'Interval', 'Note', 'Password', 'Person',
                              'Societe', 'String', 'Time', 'Workcase'])
        rconcerne = schema.rschema('concerne')
        self.assertUnorderedIterableEquals(rconcerne.subjects(), ['Workcase', 'Person'])
        self.assertUnorderedIterableEquals(rconcerne.objects(), ['Workcase', 'Societe'])
        self.assertRaises(KeyError, schema.eschema, 'Affaire')
        workcase = schema.eschema('Workcase')
        schema.__test__ = True
        self.assertEquals(workcase.rproperties('concerne'), orig_rprops)


class SymetricTC(TestCase):
    def setUp(self):
        global schema
        schema = Schema('Test Schema')
        ebug = schema.add_entity_type(EntityType('Bug'))
        estory = schema.add_entity_type(EntityType('Story'))
        eproject = schema.add_entity_type(EntityType('Project'))
        schema.add_relation_type(RelationType('see_also', symetric=True))

    def test_association_types(self):
        schema.add_relation_def(RelationDefinition('Bug', 'see_also', 'Bug'))
        schema.add_relation_def(RelationDefinition('Bug', 'see_also', 'Story'))
        schema.add_relation_def(RelationDefinition('Bug', 'see_also', 'Project'))
        schema.add_relation_def(RelationDefinition('Story', 'see_also', 'Story'))
        schema.add_relation_def(RelationDefinition('Story', 'see_also', 'Project'))
        schema.add_relation_def(RelationDefinition('Project', 'see_also', 'Project'))

        rsee_also = schema.rschema('see_also')
        subj_types = rsee_also.associations()
        subj_types.sort()
        self.assertEquals(subj_types,
                          [('Bug', ['Bug', 'Story', 'Project']),
                           ('Project', ['Bug', 'Story', 'Project']),
                           ('Story', ['Bug', 'Story', 'Project'])])

    def test_wildcard_association_types(self):
        class see_also(RelationDefinition):
            subject = '*'
            object ='*'
        see_also.expand_relation_definitions({'see_also': see_also}, schema)
        rsee_also = schema.rschema('see_also')
        subj_types = rsee_also.associations()
        subj_types.sort()
        for key, vals in subj_types:
            vals.sort()
        self.assertEquals(subj_types,
                          [('Bug', ['Bug', 'Project', 'Story']),
                           ('Project', ['Bug', 'Project', 'Story']),
                           ('Story', ['Bug', 'Project', 'Story'])])

if __name__ == '__main__':
    unittest_main()



from yams.buildobjs import EntityType, RelationType, SubjectRelation

class Anentity(EntityType):
    rel = SubjectRelation('Anentity', inlined=True)

class Anotherentity(EntityType):
    rel = SubjectRelation('Anentity')


class rel(RelationType):
    composite = 'subject'
    cardinality = '1*'
    symetric = True

class __rel(RelationType):
    name = 'rel'
    composite = 'subject'




from yams.buildobjs import EntityType, Datetime, Date, Time

class Datetest(EntityType):
    dt1 = Datetime(default=u'now')
    dt2 = Datetime(default=u'today')
    d1  = Date(default=u'today')
    d2  = Date(default=u'2007/12/11')
    t1  = Time(default=u'08:40')
    t2  = Time(default=u'09:45')




"""Object model and utilities to define generic Entities/Relations schemas.

:copyright:
  2004-2008 `LOGILAB S.A. <http://www.logilab.fr>`_ (Paris, FRANCE),
  all rights reserved.

:contact:
  http://www.logilab.org/project/yams --
  mailto:python-projects@logilab.org

:license:
  `General Public License version 2
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>`_
"""
__docformat__ = "restructuredtext en"
from yams.__pkginfo__ import version as __version__

# set _ builtin to unicode by default, should be overriden if necessary
import __builtin__
__builtin__._ = unicode

from logilab.common.compat import set

BASE_TYPES = set(('String', 'Int', 'Float', 'Boolean', 'Date', 'Decimal',
                  'Time', 'Datetime', 'Interval', 'Password', 'Bytes'))

# base groups used in permissions
BASE_GROUPS = set((_('managers'), _('users'), _('guests'), _('owners')))


from logilab.common import nullobject
MARKER = nullobject()


def use_py_datetime():
    global DATE_FACTORY_MAP, KEYWORD_MAP

    from datetime import datetime, date, time
    from time import strptime as time_strptime

    try:
        strptime = datetime.strptime
    except AttributeError: # py < 2.5
        def strptime(value, format):
            return datetime(*time_strptime(value, format)[:6])

    def strptime_time(value, format='%H:%M'):
        return time(*time_strptime(value, format)[3:6])

    KEYWORD_MAP = {'Datetime.NOW' : datetime.now,
                   'Datetime.TODAY': datetime.today,
                   'Date.TODAY': date.today}
    DATE_FACTORY_MAP = {
        'Datetime' : lambda x: ':' in x and strptime(x, '%Y/%m/%d %H:%M') or strptime(x, '%Y/%m/%d'),
        'Date' : lambda x : strptime(x, '%Y/%m/%d'),
        'Time' : strptime_time
        }

try:
    from mx.DateTime import today, now, DateTimeFrom, DateFrom, TimeFrom
    KEYWORD_MAP = {'Datetime.NOW' : now,
                   'Datetime.TODAY' : today,
                   'Date.TODAY': today}
    DATE_FACTORY_MAP = {'Datetime' : DateTimeFrom,
                        'Date' : DateFrom,
                        'Time' : TimeFrom}
except ImportError:
    use_py_datetime()

# work in progress
from yams._exceptions import *
from yams.schema import Schema, EntitySchema, RelationSchema

class _RelationRole(int):
    def __eq__(self, other):
        if isinstance(other, _RelationRole):
            return other is self
        if self:
            return other == 'object'
        return other == 'subject'
    def __nonzero__(self):
        print 'oyop'
        if self is SUBJECT:
            return OBJECT
        return SUBJECT


SUBJECT = _RelationRole(0)
OBJECT  = _RelationRole(1)

from warnings import warn

def ensure_new_subjobj(val, cls=None, attr=None):
    if isinstance(val, int):
        return val
    if val == 'subject':
        msg = 'using string instead of cubicweb.SUBJECT'
        if cls:
            msg += ' for attribute %s of class %s' % (attr, cls.__name__)
        warn(DeprecationWarning, msg)
        return SUBJECT
    if val == 'object':
        msg = 'using string instead of cubicweb.OBJECT'
        if cls:
            msg += ' for attribute %s of class %s' % (attr, cls.__name__)
        warn(DeprecationWarning, msg)
        return SUBJECT



"""Classes to define generic Entities/Relations schemas.

:organization: Logilab
:copyright: 2003-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
:license: General Public License version 2 - http://www.gnu.org/licenses
"""
__docformat__ = "restructuredtext en"

import warnings
from copy import deepcopy
from decimal import Decimal

from logilab.common.decorators import cached
from logilab.common.compat import sorted
from logilab.common.interface import implements

import yams
from yams import (BASE_TYPES, MARKER, ValidationError, BadSchemaDefinition,
                  use_py_datetime)
from yams.interfaces import (ISchema, IRelationSchema, IEntitySchema,
                             IVocabularyConstraint)
from yams.constraints import BASE_CHECKERS, BASE_CONVERTERS, UniqueConstraint

def rehash(dictionary):
    """this function manually builds a copy of `dictionary` but forces
    hash values to be recomputed. Note that dict(d) or d.copy() don't
    do that.

    It is used to :
      - circumvent Pyro / (un)pickle problems (hash mode is changed
        during reconstruction)
      - force to recompute keys' hash values. This is needed when a
        schema's type is changed because the schema's hash method is based
        on the type attribute. This problem is illusrated by the pseudo-code
        below :

        >>> topic = EntitySchema(type='Topic')
        >>> d = {topic : 'foo'}
        >>> d[topic]
        'foo'
        >>> d['Topic']
        'foo'
        >>> topic.type = 'folder'
        >>> topic in d
        False
        >>> 'Folder' in d
        False
        >>> 'Folder' in d.keys() # but it can be found "manually"
        True
        >>> d = rehash(d) # explicit rehash()
        >>> 'Folder' in d
        True
    """
    return dict(item for item in dictionary.items())

def _format_properties(props):
    res = [('%s=%s' % item) for item in props.items() if item[1]]
    return ','.join(res)


class ERSchema(object):
    """Base class shared by entity and relation schema."""

    ACTIONS = ()

    def __init__(self, schema=None, erdef=None):
        """
        Construct an ERSchema instance.

        :Parameters:
         - `schema`: (??)
         - `erdef`: (??)
        """
        if erdef is None:
            return
        assert schema
        assert erdef
        self.schema = schema
        self.type = erdef.name
        self.description = erdef.description or ''
        # mapping from action to groups
        try:
            self._groups = erdef.permissions.copy()
        except AttributeError:
            self._groups = {}

    def __cmp__(self, other):
        return cmp(self.type, getattr(other, 'type', other))

    def __hash__(self):
        try:
            if self.schema.__hashmode__ is None:
                return hash(self.type)
        except AttributeError:
            pass
        return hash(id(self))

    def __deepcopy__(self, memo):
        clone = self.__class__()
        memo[id(self)] = clone
        clone.type = deepcopy(self.type, memo)
        clone.schema = deepcopy(self.schema, memo)
        clone.schema.__hashmode__ = None
        clone.__dict__ = deepcopy(self.__dict__, memo)
        return clone

    def __str__(self):
        return self.type

    def set_groups(self, action, groups):
        """set the groups allowed to perform <action> on entities of this type

        :Parameters:
         - `action`: (str) the name of a permission
         - `groups`: (tuple) the groups with the given permission
        """
        assert type(groups) is tuple, ('groups is expected to be a tuple not %s' % type(groups))
        assert action in self.ACTIONS, ('%s not in %s' % (action, self.ACTIONS))
        self._groups[action] = groups

    def get_groups(self, action):
        """return the groups authorized to perform <action> on entities of
        this type

        :type action: str
        :param action: the name of a permission

        :rtype: tuple
        :return: the groups with the given permission
        """
        assert action in self.ACTIONS, ('%s not in %s' % (action, self.ACTIONS))
        try:
            return self._groups[action]
        except KeyError:
            return ()


    def has_group(self, action, group):
        """return true if the group is authorized for the given action

        :type action: str
        :param action: the name of a permission

        :rtype: bool
        :return: flag indicating whether the group has the permission
        """
        assert action in self.ACTIONS, ('%s not in %s' % (action, self.ACTIONS))
        return group in self._groups[action]

    def has_access(self, user, action):
        """return true if the user has the given permission on entity of this
        type

        :type user: `ginco.common.utils.User`
        :param user: a Erudi user instance

        :type action: str
        :param action: the name of a permission

        :rtype: bool
        :return: flag indicating whether the user has the permission
        """
        return user.matching_groups(self.get_groups(action))

    def set_default_groups(self):
        """set default action -> groups mapping"""
        if self._groups:
            # already initialized, check everything is fine
            for action, groups in self._groups.items():
                assert isinstance(groups, tuple), \
                       ('permission for action %s of %s isn\'t a tuple as '
                        'expected' % (action, self.type))
        else:
            self._groups = self.get_default_groups()

    def get_default_groups(self):
        """provide default action -> groups mapping"""
        raise NotImplementedError()

# Schema objects definition ###################################################

KNOWN_METAATTRIBUTES = set(('format', 'encoding'))

class EntitySchema(ERSchema):
    """An entity has a type, a set of subject and or object relations
    the entity schema defines the possible relations for a given type and some
    constraints on those relations.
    """
    __implements__ = IEntitySchema

    ACTIONS = ('read', 'add', 'update', 'delete')
    field_checkers = BASE_CHECKERS
    field_converters = BASE_CONVERTERS

    # XXX set default values for those attributes on the class level since
    # they may be missing from schemas obtained by pyro
    _specialized_type = None
    _specialized_by = []
    def __init__(self, schema=None, rdef=None, *args, **kwargs):
        super(EntitySchema, self).__init__(schema, rdef, *args, **kwargs)
        if rdef is not None:
            # quick access to bounded relation schemas
            self._subj_relations = {}
            self._obj_relations = {}
            self._specialized_type = rdef.specialized_type
            self._specialized_by = rdef.specialized_by
        else:
            self._specialized_type = None
            self._specialized_by = []

    def __repr__(self):
        return '<%s %s - %s>' % (self.type,
                                 [rs.type for rs in self.subject_relations()],
                                 [rs.type for rs in self.object_relations()])

    def _rehash(self):
        self._subj_relations = rehash(self._subj_relations)
        self._obj_relations = rehash(self._obj_relations)

    # schema building methods #################################################

    def add_subject_relation(self, rschema):
        """register the relation schema as possible subject relation"""
        self._subj_relations[rschema] = rschema

    def add_object_relation(self, rschema):
        """register the relation schema as possible object relation"""
        self._obj_relations[rschema] = rschema

    def del_subject_relation(self, rtype):
        try:
            del self._subj_relations[rtype]
        except KeyError:
            pass

    def del_object_relation(self, rtype):
        try:
            del self._obj_relations[rtype]
        except KeyError:
            pass

    def get_default_groups(self):
        """   get default action -> groups mapping  """
        if self.is_final():
            # no permissions needed for final entities, access to them
            # is defined through relations
            return {'read': ('managers', 'users', 'guests',)}
        else:
            return {'read': ('managers', 'users', 'guests',),
                    'update': ('managers', 'owners',),
                    'delete': ('managers', 'owners'),
                    'add': ('managers', 'users',)}

    # IEntitySchema interface #################################################

    def is_final(self):
        """return true if the entity is a final entity
        (and so cannot be used as subject of a relation)
        """
        return self.type in BASE_TYPES

    def is_subobject(self, strict=False):
        """return True if this entity type is contained by another. If strict,
        return True if this entity type *must* be contained by another.
        """
        for rschema in self.object_relations():
            if self.objrproperty(rschema, 'composite') == 'subject':
                if not strict:
                    return True
                if self.objrproperty(rschema, 'cardinality')[1] in '1+':
                    return True
        for rschema in self.subject_relations():
            if self.subjrproperty(rschema, 'composite') == 'object':
                if not strict:
                    return True
                if self.subjrproperty(rschema, 'cardinality')[0] in '1+':
                    return True
                return True
        return False

    def ordered_relations(self):
        """ return subject relation in an ordered way"""
        result = []
        for rschema in self._subj_relations.values():
            otype = rschema.objects(self)[0]
            result.append((rschema.rproperty(self, otype, 'order'), rschema))
        return [r[1] for r in sorted(result)]

    def subject_relations(self):
        """return a list of relations that may have this type of entity as
        subject
        """
        return self._subj_relations.values()

    def has_subject_relation(self, rtype):
        """if this entity type as a `rtype` subject relation, return its schema
        else return None
        """
        return self._subj_relations.get(rtype)

    def object_relations(self):
        """return a list of relations that may have this type of entity as
        object
        """
        return self._obj_relations.values()

    def has_object_relation(self, rtype):
        """if this entity type as a `rtype` object relation, return its schema
        else return None
        """
        return self._obj_relations.get(rtype)

    def subject_relation(self, rtype):
        """return the relation schema for the rtype subject relation

        Raise `KeyError` if rtype is not a subject relation of this entity type
        """
        return self._subj_relations[rtype]

    def object_relation(self, rtype):
        """return the relation schema for the rtype object relation

        Raise `KeyError` if rtype is not an object relation of this entity type
        """
        return self._obj_relations[rtype]

    def attribute_definitions(self):
        """return an iterator on attribute definitions

        attribute relations are a subset of subject relations where the
        object's type is a final entity

        an attribute definition is a 2-uple :
        * schema of the (final) relation
        * schema of the destination entity type
        """
        for rschema in self.ordered_relations():
            if not rschema.is_final():
                continue
            eschema = rschema.objects(self)[0]
            yield rschema, eschema

    def destination(self, rtype):
        """return the type or schema of entities related by the given relation

        `rtype` must be a subject final relation
        """
        rschema = self.subject_relation(rtype)
        assert rschema.is_final(), (self.type, rtype)
        objtypes = rschema.objects(self.type)
        assert len(objtypes) == 1
        return objtypes[0]

    def subjrproperty(self, rtype, prop):
        """convenience method to access a property of a subject relation"""
        return self.role_rproperty('subject', rtype, prop)

    def objrproperty(self, rtype, prop):
        """convenience method to access a property of an object relation"""
        return self.role_rproperty('object', rtype, prop)

    def role_rproperty(self, role, rtype, prop, ttype=None):
        """convenience method to access a property of a relation according to
        this schema role
        """
        if role == 'subject':
            rschema = self.subject_relation(rtype)
            if ttype is None:
                ttype = rschema.objects(self)[0]
            return rschema.rproperty(self, ttype, prop)
        else:
            assert role == 'object'
            rschema = self.object_relation(rtype)
            if ttype is None:
                ttype = rschema.subjects(self)[0]
            return rschema.rproperty(ttype, self, prop)

    def rproperty(self, rtype, prop):
        """convenience method to access a property of a final subject relation"""
        rschema = self.subject_relation(rtype)
        return rschema.rproperty(self, self.destination(rtype), prop)

    def set_rproperty(self, rtype, prop, value):
        """convenience method to set the value of a property of a subject relation"""
        rschema = self.subject_relation(rtype)
        return rschema.set_rproperty(self, self.destination(rtype), prop, value)

    def rproperties(self, rtype):
        """convenience method to access properties of a subject relation"""
        rschema = self.subject_relation(rtype)
        desttype = rschema.objects(self)[0]
        return rschema.rproperties(self, desttype)

    def relation_definitions(self, includefinal=False):
        """return an iterator on relation definitions

        if includefinal is false, only non attribute relation are returned

        a relation definition is a 3-uple :
        * schema of the (non final) relation
        * schemas of the possible destination entity types
        * a string telling if this is a 'subject' or 'object' relation
        """
        for rschema in self.ordered_relations():
            if includefinal or not rschema.is_final():
                yield rschema, rschema.objects(self), 'subject'
        for rschema in self.object_relations():
            yield rschema, rschema.subjects(self), 'object'

    def has_metadata(self, attr, metadata):
        """return true if this entity's schema has an encoding field for the given
        attribute
        """
        return self.has_subject_relation('%s_%s' % (attr, metadata))

    def is_metadata(self, attr):
        """return a metadata for an attribute (None if unspecified)"""
        try:
            attr, metadata = str(attr).rsplit('_', 1)
        except ValueError:
            return None
        if metadata in KNOWN_METAATTRIBUTES and self.has_subject_relation(attr):
            return (attr, metadata)
        return None

    @cached
    def meta_attributes(self):
        """return a dictionnary defining meta-attributes:
        * key is an attribute schema
        * value is a 2-uple (metadata name, described attribute name)

        a metadata attribute is expected to be named using the following scheme:

          <described attribute name>_<metadata name>

        for instance content_format is the format metadata of the content
        attribute (if it exists).
        """
        metaattrs = {}
        for rschema, _ in self.attribute_definitions():
            try:
                attr, meta = rschema.type.rsplit('_', -1)
            except ValueError:
                continue
            if meta in KNOWN_METAATTRIBUTES and self.has_subject_relation(attr):
                metaattrs[rschema] = (meta, attr)
        return metaattrs

    def main_attribute(self):
        """convenience method that returns the *main* (i.e. the first non meta)
        attribute defined in the entity schema
        """
        for rschema, _ in self.attribute_definitions():
            if not self.is_metadata(rschema):
                return rschema

    def indexable_attributes(self):
        """return the relation schema of attribtues to index"""
        assert not self.is_final()
        for rschema in self.subject_relations():
            if rschema.is_final():
                if self.rproperty(rschema, 'fulltextindexed'):
                    yield rschema

    def fulltext_relations(self):
        """return the (name, role) of relations to index"""
        assert not self.is_final()
        for rschema in self.subject_relations():
            if not rschema.is_final() and  rschema.fulltext_container == 'subject':
                yield rschema, 'subject'
        for rschema in self.object_relations():
            if rschema.fulltext_container == 'object':
                yield rschema, 'object'

    def fulltext_containers(self):
        """return relations whose extremity points to an entity that should
        contains the full text index content of entities of this type
        """
        for rschema in self.subject_relations():
            if rschema.fulltext_container == 'object':
                yield rschema, 'object'
        for rschema in self.object_relations():
            if rschema.fulltext_container == 'subject':
                yield rschema, 'subject'

    def defaults(self):
        """return an iterator on (attribute name, default value)"""
        assert not self.is_final()
        for rschema in self.subject_relations():
            if rschema.is_final():
                value = self.default(rschema)
                if value is not None:
                    yield rschema, value

    def default(self, rtype):
        """return the default value of a subject relation"""
        default =  self.rproperty(rtype, 'default')
        if callable(default):
            default = default()
        if default is MARKER:
            default = None
        if default is not None:
            attrtype = self.destination(rtype)
            if attrtype == 'Boolean':
                if not isinstance(default, bool):
                    default = default == 'True'
            elif attrtype == 'Int':
                if not isinstance(default, int):
                    default = int(default)
            elif attrtype == 'Float':
                if not isinstance(default, float):
                    default = float(default)
            elif attrtype == 'Decimal':
                if not isinstance(default, Decimal):
                    default = Decimal(default)
            elif attrtype in ('Date', 'Datetime', 'Time'):
                try:
                    default = yams.KEYWORD_MAP['%s.%s' % (attrtype, default.upper())]()
                except KeyError:
                    default = yams.DATE_FACTORY_MAP[attrtype](default)
            else:
                default = unicode(default)
        return default

    def has_unique_values(self, rtype):
        """convenience method to check presence of the UniqueConstraint on a
        relation
        """
        rschema = self.subject_relation(rtype)
        for constraint in self.constraints(rtype):
            if isinstance(constraint, UniqueConstraint):
                return True
        return False

    def constraints(self, rtype):
        """return constraint of type <cstrtype> associated to the <rtype>
        subjet relation

        returns None if no constraint of type <cstrtype> is found
        """
        return self.rproperty(rtype, 'constraints')

    def vocabulary(self, rtype):
        """backward compat return the vocabulary of a subject relation
        """
        for constraint in self.constraints(rtype):
            if implements(constraint, IVocabularyConstraint):
                break
        else:
            raise AssertionError('field %s of entity %s has no vocabulary' %
                                 (rtype, self))
        return constraint.vocabulary()

    def cardinality(self, rtype, role):
        """return cardinality (a single letter in 1?*+) for the given relation"""
        rschema = self.schema.rschema(rtype)
        if role == 'subject':
            targetschema = rschema.objects(self)[0]
            return rschema.rproperty(self, targetschema, 'cardinality')[0]
        targetschema = rschema.subjects(self)[0]
        return rschema.rproperty(targetschema, self, 'cardinality')[1]

    def check(self, entity, creation=False, _=unicode):
        """check the entity and raises an ValidationError exception if it
        contains some invalid fields (ie some constraints failed)
        """
        assert not self.is_final()
        errors = {}
        for rschema in self.subject_relations():
            if not rschema.is_final():
                continue
            aschema = self.destination(rschema)
            # don't care about rhs cardinality, always '*' (if it make senses)
            card = rschema.rproperty(self, aschema, 'cardinality')[0]
            assert card in '?1'
            required = card == '1'
            # check value according to their type
            try:
                value = entity[rschema]
            except KeyError:
                if creation and required:
                    # missing required attribute with no default on creation
                    # is not autorized
                    errors[rschema.type] = _('required attribute')
                # on edition, missing attribute is considered as no changes
                continue
            # skip other constraint if value is None and None is allowed
            if value is None:
                if required:
                    errors[rschema.type] = _('required attribute')
                continue
            if not aschema.check_value(value):
                errors[rschema.type] = _('incorrect value (%(value)s) for type "%(type)s"') % {
                    'value':value, 'type': _(aschema.type)}
                if isinstance(value, str):
                    errors[rschema.type] = '%s; you might want to try unicode'  % errors[rschema.type]
                continue
            # ensure value has the correct python type
            entity[rschema] = value = aschema.convert_value(value)
            # check arbitrary constraints
            for constraint in rschema.rproperty(self, aschema, 'constraints'):
                if not constraint.check(entity, rschema, value):
                    errors[rschema.type] = _('%(cstr)s constraint failed for value %(value)r') % {
                        'cstr': constraint, 'value': value}
        if errors:
            raise ValidationError(entity, errors)

    def check_value(self, value):
        """check the value of a final entity (ie a const value)"""
        assert self.is_final()
        return self.field_checkers[self](self, value)

    def convert_value(self, value):
        """check the value of a final entity (ie a const value)"""
        assert self.is_final()
        try:
            return self.field_converters[self](value)
        except KeyError:
            return value

    def specializes(self):
        if self._specialized_type:
            return self.schema.eschema(self._specialized_type)
        return None

    def ancestors(self):
        specializes = self.specializes()
        ancestors = []
        while specializes:
            ancestors.append(specializes)
            specializes = specializes.specializes()
        return ancestors

    def specialized_by(self, recursive=True):
        eschema = self.schema.eschema
        subschemas = [eschema(etype) for etype in self._specialized_by]
        if recursive:
            for subschema in subschemas[:]:
                subschemas.extend(subschema.specialized_by(recursive=True))
        return subschemas

    # bw compat
    subject_relation_schema = subject_relation
    object_relation_schema = object_relation


class RelationSchema(ERSchema):
    """A relation is a named and oriented link between two entities.
    A relation schema defines the possible types of both extremities.

    Cardinality between the two given entity's type is defined
    as a 2 characters string where each character is one of:
     - 1 <-> 1..1 <-> one and only one
     - ? <-> 0..1 <-> zero or one
     - + <-> 1..n <-> one or more
     - * <-> 0..n <-> zero or more
    """
    ACTIONS = ('read', 'add', 'delete')
    _RPROPERTIES = {'cardinality': None,
                    'constraints': (),
                    'order': 9999,
                    'description': '',
                    'infered': False,}
    _NONFINAL_RPROPERTIES = {'composite': None}
    _FINAL_RPROPERTIES = {'default': None,
                          'uid': False,
                          'indexed': False}
    _STRING_RPROPERTIES = {'fulltextindexed': False,
                           'internationalizable': False}
    _BYTES_RPROPERTIES = {'fulltextindexed': False}

    __implements__ = IRelationSchema

    def __init__(self, schema=None, rdef=None, **kwargs):
        if rdef is not None:
            # if this relation is symetric/inlined
            self.symetric = rdef.symetric or False
            self.inlined = rdef.inlined or False
            # if full text content of subject/object entity should be added
            # to other side entity (the container)
            self.fulltext_container = rdef.fulltext_container or None
            # if this relation is an attribute relation
            self.final = False
            # mapping to subject/object with schema as key
            self._subj_schemas = {}
            self._obj_schemas = {}
            # relation properties
            self._rproperties = {}
        super(RelationSchema, self).__init__(schema, rdef, **kwargs)

    def __repr__(self):
        return '<%s [%s]>' % (self.type,
                              '; '.join('%s,%s:%s'%(s.type, o.type, _format_properties(props))
                                        for (s, o), props in self._rproperties.items()))

    def _rehash(self):
        self._subj_schemas = rehash(self._subj_schemas)
        self._obj_schemas = rehash(self._obj_schemas)
        self._rproperties = rehash(self._rproperties)

    # schema building methods #################################################

    def update(self, subjschema, objschema, rdef):
        """Allow this relation between the two given types schema"""
        if subjschema.is_final():
            msg = 'type %s can\'t be used as subject in a relation' % subjschema
            raise BadSchemaDefinition(msg)
        # check final consistency:
        # * a final relation only points to final entity types
        # * a non final relation only points to non final entity types
        final = objschema.is_final()
        for eschema in self.objects():
            if eschema is objschema:
                continue
            if final != eschema.is_final():
                if final:
                    frschema, nfrschema = objschema, eschema
                else:
                    frschema, nfrschema = eschema, objschema
                msg = "ambiguous relation %s: %s is final but not %s" % (
                    self.type, frschema, nfrschema)
                raise BadSchemaDefinition(msg)
        constraints = getattr(rdef, 'constraints', None)
        if constraints:
            for cstr in constraints:
                cstr.check_consistency(subjschema, objschema, rdef)
        self.final = final
        # update our internal struct
        self._update(subjschema, objschema)
        if self.symetric:
            self._update(objschema, subjschema)
            try:
                self.init_rproperties(subjschema, objschema, rdef)
                if objschema != subjschema:
                    self.init_rproperties(objschema, subjschema, rdef)
            except BadSchemaDefinition:
                # this is authorized for consistency
                pass
        else:
            self.init_rproperties(subjschema, objschema, rdef)
        if self.inlined and self.rproperty(subjschema, objschema, 'cardinality')[0] in '*+':
            raise BadSchemaDefinition('inlined relation %s can\'t have multiple cardinality for its subject %s'
                                      % (self.type, subjschema))
        # update extremities schema
        subjschema.add_subject_relation(self)
        if self.symetric:
            objschema.add_subject_relation(self)
        else:
            objschema.add_object_relation(self)

    def _update(self, subjectschema, objectschema):
        objtypes = self._subj_schemas.setdefault(subjectschema, [])
        if not objectschema in objtypes:
            objtypes.append(objectschema)
        subjtypes = self._obj_schemas.setdefault(objectschema, [])
        if not subjectschema in subjtypes:
            subjtypes.append(subjectschema)

    def del_relation_def(self, subjschema, objschema, _recursing=False):
        try:
            self._subj_schemas[subjschema].remove(objschema)
            if len(self._subj_schemas[subjschema]) == 0:
                del self._subj_schemas[subjschema]
                subjschema.del_subject_relation(self)
        except (ValueError, KeyError):
            pass
        try:
            self._obj_schemas[objschema].remove(subjschema)
            if len(self._obj_schemas[objschema]) == 0:
                del self._obj_schemas[objschema]
                objschema.del_object_relation(self)
        except (ValueError, KeyError):
            pass
        try:
            del self._rproperties[(subjschema, objschema)]
        except KeyError:
            pass
        try:
            if self.symetric and subjschema != objschema and not _recursing:
                self.del_relation_def(objschema, subjschema, True)
        except KeyError:
            pass
        if not self._obj_schemas or not self._subj_schemas:
            assert not self._obj_schemas and not self._subj_schemas
            return True
        return False

    def get_default_groups(self):
        if self.final:
            return {'read': ('managers', 'users', 'guests'),
                    'delete': ('managers', 'users', 'guests'),
                    'add': ('managers', 'users', 'guests')}
        else:
            return {'read': ('managers', 'users', 'guests',),
                    'delete': ('managers', 'users'),
                    'add': ('managers', 'users',)}

    # relation definitions properties handling ################################

    def rproperty_defs(self, desttype):
        """return a dictionary mapping property name to its definition
        for each allowable properties when the relation has `desttype` as
        target entity's type
        """
        propdefs = self._RPROPERTIES.copy()
        if not self.is_final():
            propdefs.update(self._NONFINAL_RPROPERTIES)
        else:
            propdefs.update(self._FINAL_RPROPERTIES)
            if desttype == 'String':
                propdefs.update(self._STRING_RPROPERTIES)
            elif desttype == 'Bytes':
                propdefs.update(self._BYTES_RPROPERTIES)
        return propdefs

    def iter_rdefs(self):
        """return an iterator on (subject, object) of this relation"""
        return self._rproperties.iterkeys()

    def rdefs(self):
        """return a list of (subject, object) of this relation"""
        return self._rproperties.keys()

    def has_rdef(self, subj, obj):
        return (subj, obj) in self._rproperties

    def rproperties(self, subject, object):
        """return the properties dictionary of a relation"""
        try:
            return self._rproperties[(subject, object)]
        except KeyError:
            raise KeyError('%s %s %s' % (subject, self, object))

    def rproperty(self, subject, object, property):
        """return the property for a relation definition"""
        return self.rproperties(subject, object).get(property)

    def set_rproperty(self, subject, object, pname, value):
        """set value for a subject relation specific property"""
        assert pname in self.rproperty_defs(object)
        self._rproperties[(subject, object)][pname] = value

    def init_rproperties(self, subject, object, rdef):
        key = subject, object
        if key in self._rproperties:
            msg = '(%s, %s) already defined for %s' % (subject, object, self)
            raise BadSchemaDefinition(msg)
        self._rproperties[key] = {}
        for prop, default in self.rproperty_defs(key[1]).iteritems():
            rdefval = getattr(rdef, prop, MARKER)
            if rdefval is MARKER:
                if prop == 'cardinality':
                    default = (object in BASE_TYPES) and '?1' or '**'
            else:
                default = rdefval
            self._rproperties[key][prop] = default

    # IRelationSchema interface ###############################################

    def is_final(self):
        """return true if this relation has final object entity's types

        (we enforce that a relation can't point to both final and non final
        entity's type)
        """
        return self.final

    def associations(self):
        """return a list of (subject, [objects]) defining between
        which types this relation may exists
        """
        # XXX deprecates in favor of iter_rdefs() ?
        return self._subj_schemas.items()

    def subjects(self, etype=None):
        """Return a list of entity schemas which can be subject of this relation.

        If etype is not None, return a list of schemas which can be subject of
        this relation with etype as object.

        :raise `KeyError`: if etype is not a subject entity type.
        """
        if etype is None:
            return tuple(self._subj_schemas.keys())
        try:
            return tuple(self._obj_schemas[etype])
        except KeyError:
            raise KeyError("%s don't have %s as object" % (self, etype))

    def objects(self, etype=None):
        """Return a list of entity schema which can be object of this relation.

        If etype is not None, return a list of schemas which can be object of
        this relation with etype as subject.

        :raise `KeyError`: if etype is not an object entity type.
        """
        if etype is None:
            return tuple(self._obj_schemas.keys())
        try:
            return tuple(self._subj_schemas[etype])
        except KeyError:
            raise KeyError("%s don't have %s as subject" % (self, etype))

    def targets(self, etype, role='subject'):
        """return possible target types with <etype> as <x>"""
        assert role in ('subject', 'object')
        if role == 'subject':
            return self.objects(etype)
        return self.subjects(etype)

    def constraint_by_type(self, subjtype, objtype, cstrtype):
        for cstr in self.rproperty(subjtype, objtype, 'constraints'):
            if cstr.type() == cstrtype:
                return cstr
        return None


class Schema(object):
    """set of entities and relations schema defining the possible data sets
    used in an application


    :type name: str
    :ivar name: name of the schema, usually the application identifier

    :type base: str
    :ivar base: path of the directory where the schema is defined
    """
    __implements__ = ISchema
    entity_class = EntitySchema
    relation_class = RelationSchema
    # __hashmode__ is a evil hack to support schema pickling
    # it should be set to 'pickle' before pickling is done and reset to None
    # once it's done
    __hashmode__ = 'pickle' # None | 'pickle'

    def __init__(self, name, construction_mode='strict'):
        super(Schema, self).__init__()
        self.__hashmode__ = None
        self.name = name
        # with construction_mode != 'strict', no error when trying to add a
        # relation using an undefined entity type, simply log the error
        self.construction_mode = construction_mode
        self._entities = {}
        self._relations = {}

    def __setstate__(self, state):
        self.__dict__.update(state)
        # restore __hashmode__
        self.__hashmode__ = None
        self._rehash()

    def _rehash(self):
        """rehash schema's internal structures"""
        for eschema in self._entities.values():
            eschema._rehash()
        for rschema in self._relations.values():
            rschema._rehash()

    def get(self, name, default=None):
        try:
            return self[name]
        except KeyError:
            return default

    def __getitem__(self, name):
        try:
            return self.eschema(name)
        except KeyError:
            return self.rschema(name)

    def __contains__(self, name):
        try:
            self[name]
        except KeyError:
            return False
        return True

    # schema building methods #################################################

    def add_entity_type(self, edef):
        """Add an entity schema definition for an entity's type.

        :type edef: str
        :param edef: the name of the entity type to define

        :raise `BadSchemaDefinition`: if the entity type is already defined
        :rtype: `EntitySchema`
        :return: the newly created entity schema instance
        """
        etype = edef.name
        if etype in self._entities:
            msg = "entity type %s is already defined" % etype
            raise BadSchemaDefinition(msg)
        eschema = self.entity_class(self, edef)
        self._entities[etype] = eschema
        return eschema

    def rename_entity_type(self, oldname, newname):
        """renames an entity type and update internal structures accordingly
        """
        assert oldname in self._entities
        eschema = self._entities.pop(oldname)
        eschema.type = newname
        self._entities[newname] = eschema
        # rebuild internal structures since eschema's hash value has changed
        self._rehash()


    def add_relation_type(self, rtypedef):
        rtype = rtypedef.name
        if rtype in self._relations:
            msg = "relation type %s is already defined" % rtype
            raise BadSchemaDefinition(msg)
        rschema = self.relation_class(self, rtypedef)
        self._relations[rtype] = rschema
        return rschema

    def add_relation_def(self, rdef):
        """build a part of a relation schema:
        add a relation between two specific entity's types

        :rtype: RelationSchema
        :return: the newly created or simply completed relation schema
        """
        rtype = rdef.name
        try:
            rschema = self.rschema(rtype)
        except KeyError:
            return self._building_error('using unknown relation type in %s',
                                        rdef)
        try:
            subjectschema = self.eschema(rdef.subject)
        except KeyError:
            return self._building_error('using unknown type %r in relation %s',
                                        rdef.subject, rtype)
        try:
            objectschema = self.eschema(rdef.object)
        except KeyError:
            return self._building_error("using unknown type %r in relation %s",
                                        rdef.object, rtype)
        rschema.update(subjectschema, objectschema, rdef)
        return True

    def _building_error(self, msg, *args):
        if self.construction_mode == 'strict':
            raise BadSchemaDefinition(msg % args)
        self.critical(msg, *args)

    def del_relation_def(self, subjtype, rtype, objtype):
        subjschema = self.eschema(subjtype)
        objschema = self.eschema(objtype)
        rschema = self.rschema(rtype)
        if rschema.del_relation_def(subjschema, objschema):
            del self._relations[rtype]

    def del_relation_type(self, rtype):
        # XXX don't iter directly on the dictionary since it may be changed
        # by del_relation_def
        for subjtype, objtype in self.rschema(rtype)._rproperties.keys():
            self.del_relation_def(subjtype, rtype, objtype)
        if not self.rschema(rtype)._rproperties:
            del self._relations[rtype]

    def del_entity_type(self, etype):
        eschema = self._entities[etype]
        for rschema in eschema._subj_relations.values():
            for objtype in rschema.objects(etype):
                self.del_relation_def(eschema, rschema, objtype)
        for rschema in eschema._obj_relations.values():
            for subjtype in rschema.subjects(etype):
                self.del_relation_def(subjtype, rschema, eschema)
        del self._entities[etype]

    def infer_specialization_rules(self):
        # XXX
        class XXXRelationDef:
            infered = True
        for rschema in self.relations():
            if rschema.is_final():
                continue
            for subject, object in rschema.rdefs():
                subjeschemas = [subject] + subject.specialized_by(recursive=True)
                objeschemas = [object] + object.specialized_by(recursive=True)
                for subjschema in subjeschemas:
                    for objschema in objeschemas:
                        subjobj = (subjschema, objschema)
                        # don't try to add an already defined relation
                        if subjobj in rschema.rdefs():
                            continue
                        rschema.update(subjschema, objschema, XXXRelationDef)

    def remove_infered_definitions(self):
        """remove any infered definitions added by
        `infer_specialization_rules`
        """
        for rschema in self.relations():
            if rschema.is_final():
                continue
            for subject, object in rschema.rdefs():
                if rschema.rproperty(subject, object, 'infered'):
                    self.del_relation_def(subject, rschema, object)

    def rebuild_infered_relations(self):
        """remove any infered definitions and rebuild them"""
        self.remove_infered_definitions()
        self.infer_specialization_rules()

    # ISchema interface #######################################################

    def entities(self):
        """return a list of possible entity's type

        :rtype: list
        :return: defined entity's types (str) or schemas (`EntitySchema`)
        """
        return self._entities.values()

    def has_entity(self, etype):
        """return true the type is defined in the schema

        :type etype: str
        :param etype: the entity's type

        :rtype: bool
        :return:
          a boolean indicating whether this type is defined in this schema
        """
        return etype in self._entities

    def eschema(self, etype):
        """return the entity's schema for the given type

        :rtype: `EntitySchema`
        :raise `KeyError`: if the type is not defined as an entity
        """
        try:
            return self._entities[etype]
        except KeyError:
            if isinstance(etype, tuple):
                etype = list(etype)
            raise KeyError('No entity named %s in schema' % etype)

    def relations(self):
        """return the list of possible relation'types

        :rtype: list
        :return: defined relation's types (str) or schemas (`RelationSchema`)
        """
        return self._relations.values()

    def has_relation(self, rtype):
        """return true the relation is defined in the schema

        :type rtype: str
        :param rtype: the relation's type

        :rtype: bool
        :return:
          a boolean indicating whether this type is defined in this schema
        """
        return rtype in self._relations

    def rschema(self, rtype):
        """return the relation schema for the given type

        :rtype: `RelationSchema`
        """
        try:
            return self._relations[rtype]
        except KeyError:
            raise KeyError('No relation named %s in schema'%rtype)

    def final_relations(self):
        """return the list of possible final relation'types

        :rtype: list
        :return: defined relation's types (str) or schemas (`RelationSchema`)
        """
        for rschema in self.relations():
            if rschema.is_final():
                if schema:
                    yield rschema
                else:
                    yield rschema.type

    def nonfinal_relations(self):
        """return the list of possible final relation'types

        :rtype: list
        :return: defined relation's types (str) or schemas (`RelationSchema`)
        """
        for rschema in self.relations():
            if not rschema.is_final():
                if schema:
                    yield rschema
                else:
                    yield rschema.type

    # bw compat
    relation_schema = rschema
    entity_schema = eschema


import logging
from logilab.common.logging_ext import set_log_methods
LOGGER = logging.getLogger('yams')
set_log_methods(Schema, LOGGER)



"""specialization tests
"""
from logilab.common.testlib import TestCase, unittest_main

from yams.reader import SchemaLoader



class SpecializationTC(TestCase):
    def setUp(self):
        SchemaLoader.main_schema_directory = 'spschema'
        self.schema = SchemaLoader().load([self.datadir], 'Test')

    def tearDown(self):
        SchemaLoader.main_schema_directory = 'schema'

    def test_schema_specialization(self):
        schema = self.schema
        # company
        company = schema.eschema('Company')
        self.assertEquals(company.specializes(), None)
        # division
        division = schema.eschema('Division')
        self.assertEquals(division.specializes().type, 'Company')
        # subdivision
        subdivision = schema.eschema('SubDivision')
        self.assertEquals(subdivision.specializes().type, 'Division')
        # subsubdivision
        subsubdivision = schema.eschema('SubSubDivision')
        self.assertEquals(subsubdivision.specializes(), None)

    def test_ancestors(self):
        schema = self.schema
        # company
        company = schema.eschema('Company')
        self.assertEquals(company.ancestors(), [])
        # division
        division = schema.eschema('Division')
        self.assertEquals(division.ancestors(), ['Company'])
        # subdivision
        subdivision = schema.eschema('SubDivision')
        self.assertEquals(subdivision.ancestors(), ['Division', 'Company'])
        # subsubdivision
        subsubdivision = schema.eschema('SubSubDivision')
        self.assertEquals(subsubdivision.ancestors(), [])

    def test_specialized_by(self):
        schema = self.schema
        # company
        company = schema.eschema('Company')
        self.assertEquals(sorted(company.specialized_by(False)), ['Division', 'SubCompany'])
        self.assertEquals(sorted(company.specialized_by(True)), ['Division', 'SubCompany', 'SubDivision'])
        # division
        division = schema.eschema('Division')
        self.assertEquals(sorted(division.specialized_by(False)), ['SubDivision'])
        self.assertEquals(sorted(division.specialized_by(True)), ['SubDivision'])
        # subdivision
        subdivision = schema.eschema('SubDivision')
        self.assertEquals(sorted(subdivision.specialized_by(False)), [])
        # subsubdivision
        subsubdivision = schema.eschema('SubSubDivision')
        self.assertEquals(subsubdivision.specialized_by(False), [])

    def test_relations_infered(self):
        entities = [str(e) for e in self.schema.entities() if not e.is_final()]
        relations = sorted([r for r in self.schema.relations() if not r.final])
        self.assertListEquals(sorted(entities), ['Company', 'Division', 'Person',
                                                 'Student', 'SubCompany', 'SubDivision', 'SubSubDivision'])
        self.assertListEquals(relations, ['division_of', 'knows', 'works_for'])
        expected = {('Person', 'Person'): False,
                    ('Person', 'Student'): True,
                    # as Student extends Person, it already has the `knows` relation
                    ('Student', 'Person'): False,
                    ('Student', 'Student'): True,
                    }
        done = set()
        drschema, krschema, wrschema = relations
        for subjobj in krschema.rdefs():
            subject, object = subjobj
            done.add(subjobj)
            self.failUnless(subjobj in expected)
            self.assertEquals(krschema.rproperty(subject, object, 'infered'),
                              expected[subjobj])
        self.assertEquals(len(set(expected) - done), 0, 'missing %s' % (set(expected) - done))
        expected = {('Person', 'Company'): False,
                    ('Person', 'Division'): True,
                    ('Person', 'SubDivision'): True,
                    ('Person', 'SubCompany'): True,
                    ('Student', 'Company'): False,
                    ('Student', 'Division'): True,
                    ('Student', 'SubDivision'): True,
                    ('Student', 'SubCompany'): True,
                    }
        done = set()
        for subjobj in wrschema.rdefs():
            subject, object = subjobj
            done.add(subjobj)
            self.failUnless(subjobj in expected)
            self.assertEquals(wrschema.rproperty(subject, object, 'infered'),
                              expected[subjobj])
        self.assertEquals(len(set(expected) - done), 0, 'missing %s' % (set(expected) - done))

    def test_remove_infered_relations(self):
        self.schema.remove_infered_definitions()
        relations = sorted([r for r in self.schema.relations() if not r.final])
        self.assertListEquals(relations, ['division_of', 'knows', 'works_for'])
        expected = {('Person', 'Person'): False,
                    # as Student extends Person, it already has the `knows` relation
                    ('Student', 'Person'): False,
                    }
        done = set()
        drschema, krschema, wrschema = relations
        for subjobj in krschema.rdefs():
            subject, object = subjobj
            done.add(subjobj)
            self.failUnless(subjobj in expected)
            self.assertEquals(krschema.rproperty(subject, object, 'infered'),
                              expected[subjobj])
        self.assertEquals(len(set(expected) - done), 0, 'missing %s' % (set(expected) - done))
        expected = {('Person', 'Company'): False,
                    ('Student', 'Company'): False,
                   }
        done = set()
        for subjobj in wrschema.rdefs():
            subject, object = subjobj
            done.add(subjobj)
            self.failUnless(subjobj in expected)
            self.assertEquals(wrschema.rproperty(subject, object, 'infered'),
                              expected[subjobj])
        self.assertEquals(len(set(expected) - done), 0, 'missing %s' % (set(expected) - done))


if __name__ == '__main__':
    unittest_main()




from yams.buildobjs import EntityType, SubjectRelation, String

class Person(EntityType):
    firstname = String()
    knows = SubjectRelation('Person')
    works_for = SubjectRelation('Company')

class Student(Person):
    __specializes_schema__ = True

class Company(EntityType):
    name = String()

class SubCompany(Company):
    __specializes_schema__ = True

class Division(Company):
    __specializes_schema__ = True
    division_of = SubjectRelation('Company')

class SubDivision(Division):
    __specializes_schema__ = True

# This class doesn't extend the schema
class SubSubDivision(SubDivision):
    pass




"""write a schema as sql

:organization: Logilab
:copyright: 2004-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
:license: General Public License version 2 - http://www.gnu.org/licenses
"""
__docformat__ = "restructuredtext en"
__metaclass__ = type

from logilab.common.compat import sorted

from yams.constraints import SizeConstraint, UniqueConstraint

# default are usually not handled at the sql level. If you want them, set
# SET_DEFAULT to True
SET_DEFAULT = False


def schema2sql(dbhelper, schema, skip_entities=(), skip_relations=(), prefix=''):
    """write to the output stream a SQL schema to store the objects
    corresponding to the given schema
    """
    output = []
    w = output.append
    for etype in sorted(schema.entities()):
        eschema = schema.eschema(etype)
        if eschema.is_final() or eschema.type in skip_entities:
            continue
        w(eschema2sql(dbhelper, eschema, skip_relations, prefix=prefix))
    for rtype in sorted(schema.relations()):
        rschema = schema.rschema(rtype)
        if rschema.is_final() or rschema.inlined:
            continue
        w(rschema2sql(rschema))
    return '\n'.join(output)


def dropschema2sql(schema, skip_entities=(), skip_relations=(), prefix=''):
    """write to the output stream a SQL schema to store the objects
    corresponding to the given schema
    """
    output = []
    w = output.append
    for etype in sorted(schema.entities()):
        eschema = schema.eschema(etype)
        if eschema.is_final() or eschema.type in skip_entities:
            continue
        w(dropeschema2sql(eschema, skip_relations, prefix=prefix))
    for rtype in sorted(schema.relations()):
        rschema = schema.rschema(rtype)
        if rschema.is_final() or rschema.inlined:
            continue
        w(droprschema2sql(rschema))
    return '\n'.join(output)


def eschema_attrs(eschema, skip_relations):
    attrs = [attrdef for attrdef in eschema.attribute_definitions()
             if not attrdef[0].type in skip_relations]
    attrs += [(rschema, None)
              for rschema in eschema.subject_relations()
              if not rschema.final and rschema.inlined]
    return attrs


def dropeschema2sql(eschema, skip_relations=(), prefix=''):
    """return sql to drop an entity type's table"""
    # not necessary to drop indexes, that's implictly done when dropping
    # the table
    return 'DROP TABLE %s;' % eschema.type


def eschema2sql(dbhelper, eschema, skip_relations=(), prefix=''):
    """write an entity schema as SQL statements to stdout"""
    output = []
    w = output.append
    table = prefix + eschema.type
    w('CREATE TABLE %s(' % (table))
    attrs = eschema_attrs(eschema, skip_relations)
    # XXX handle objectinline physical mode
    for i in xrange(len(attrs)):
        rschema, attrschema = attrs[i]
        if attrschema is not None:
            sqltype = aschema2sql(dbhelper, eschema, rschema, attrschema,
                                  indent=' ')
        else: # inline relation
            # XXX integer is ginco specific
            sqltype = 'integer'
        if i == len(attrs) - 1:
            w(' %s%s %s' % (prefix, rschema.type, sqltype))
        else:
            w(' %s%s %s,' % (prefix, rschema.type, sqltype))
    w(');')
    # create index
    for i in xrange(len(attrs)):
        rschema, attrschema = attrs[i]
        if attrschema is None or eschema.rproperty(rschema.type, 'indexed'):
            w(dbhelper.sql_create_index(table, prefix + rschema.type))
    w('')
    return '\n'.join(output)


def aschema2sql(dbhelper, eschema, rschema, aschema, creating=True, indent=''):
    """write an attribute schema as SQL statements to stdout"""
    attr = rschema.type
    constraints = rschema.rproperty(eschema.type, aschema.type, 'constraints')
    sqltype = type_from_constraints(dbhelper, aschema.type, constraints,
                                    creating)
    if SET_DEFAULT:
        default = eschema.default(attr)
        if default is not None:
            if aschema.type == 'Boolean':
                sqltype += ' DEFAULT %s' % dbhelper.boolean_value(default)
            elif aschema.type == 'String':
                sqltype += ' DEFAULT %r' % str(default)
            elif aschema.type in ('Int', 'Float'):
                sqltype += ' DEFAULT %s' % default
            # XXX ignore default for other type
            # this is expected for NOW / TODAY
    if creating:
        if eschema.rproperty(attr, 'uid'):
            sqltype += ' PRIMARY KEY'
        elif rschema.rproperty(eschema.type, aschema.type, 'cardinality')[0] == '1':
            # don't set NOT NULL if backend isn't able to change it later
            if dbhelper.alter_column_support:
                sqltype += ' NOT NULL'
    # else we're getting sql type to alter a column, we don't want key / indexes
    # / null modifiers
    return sqltype


def type_from_constraints(dbhelper, etype, constraints, creating=True):
    """return a sql type string corresponding to the constraints"""
    constraints = list(constraints)
    unique, sqltype = False, None
    if etype == 'String':
        for constraint in constraints:
            if isinstance(constraint, SizeConstraint):
                if constraint.max is not None:
                    sqltype = 'varchar(%s)' % constraint.max
            elif isinstance(constraint, UniqueConstraint):
                unique = True
    if sqltype is None:
        sqltype = dbhelper.TYPE_MAPPING[etype]
    if creating and unique:
        sqltype += ' UNIQUE'
    return sqltype

# backward compat
_type_from_constraints = type_from_constraints


_SQL_SCHEMA = """
CREATE TABLE %(table)s (
  eid_from INTEGER NOT NULL,
  eid_to INTEGER NOT NULL,
  CONSTRAINT %(table)s_p_key PRIMARY KEY(eid_from, eid_to)
);

CREATE INDEX %(table)s_from_idx ON %(table)s(eid_from);
CREATE INDEX %(table)s_to_idx ON %(table)s(eid_to);"""


def rschema2sql(rschema):
    return _SQL_SCHEMA % {'table': '%s_relation' % rschema.type}


def droprschema2sql(rschema):
    """return sql to drop a relation type's table"""
    # not necessary to drop indexes, that's implictly done when dropping
    # the table
    return 'DROP TABLE %s_relation;' % rschema.type


def grant_schema(schema, user, set_owner=True, skip_entities=(), prefix=''):
    """write to the output stream a SQL schema to store the objects
    corresponding to the given schema
    """
    output = []
    w = output.append
    for etype in sorted(schema.entities()):
        eschema = schema.eschema(etype)
        if eschema.is_final() or etype in skip_entities:
            continue
        w(grant_eschema(eschema, user, set_owner, prefix=prefix))
    for rtype in sorted(schema.relations()):
        rschema = schema.rschema(rtype)
        if rschema.is_final() or rschema.inlined:
            continue
        w(grant_rschema(rschema, user, set_owner))
    return '\n'.join(output)


def grant_eschema(eschema, user, set_owner=True, prefix=''):
    output = []
    w = output.append
    etype = eschema.type
    if set_owner:
        w('ALTER TABLE %s%s OWNER TO %s;' % (prefix, etype, user))
    w('GRANT ALL ON %s%s TO %s;' % (prefix, etype, user))
    return '\n'.join(output)


def grant_rschema(rschema, user, set_owner=True):
    output = []
    if set_owner:
        output.append('ALTER TABLE %s_relation OWNER TO %s;' % (rschema.type, user))
    output.append('GRANT ALL ON %s_relation TO %s;' % (rschema.type, user))
    return '\n'.join(output)



from yams.buildobjs import EntityType, RelationType, SubjectRelation, \
     ObjectRelation, Int, String,  Boolean
from yams.constraints import SizeConstraint, UniqueConstraint

from data.schema import RESTRICTED_RTYPE_PERMS

class State(EntityType):
    """used to associate simple states to an entity
    type and/or to define workflows
    """
    permissions = {
        'read':   ('managers', 'users', 'guests',),
        'add':    ('managers', 'users',),
        'delete': ('managers', 'owners',),
        'update': ('managers', 'owners',),
        }

    # attributes
    eid = Int(required=True, uid=True)
    name = String(required=True,
                  indexed=True, internationalizable=True,
                  constraints=[SizeConstraint(256)])
    description = String(fulltextindexed=True, meta=True) # XXX meta to test bw compat
    # relations
    state_of = SubjectRelation('Eetype', cardinality='+*')
    next_state = SubjectRelation('State', cardinality='**')
    initial_state = ObjectRelation('Eetype', cardinality='?*')


class state_of(RelationType):
    """link a state to one or more entity type"""
    permissions = RESTRICTED_RTYPE_PERMS

class next_state(RelationType):
    """define a workflow by associating a state to possible following states
    """
    permissions = RESTRICTED_RTYPE_PERMS

class initial_state(RelationType):
    """indicate which state should be used by default when an entity using states
    is created
    """
    permissions = {
        'read':   ('managers', 'users', 'guests',),
        'add':    ('managers', 'users',),
        'delete': ('managers', 'users',),
        }
    inlined = True

class Eetype(EntityType):
    """define an entity type, used to build the application schema"""
    permissions = {
        'read':   ('managers', 'users', 'guests',),
        'add':    ('managers',),
        'delete': ('managers',),
        'update': ('managers', 'owners',),
        }
    name = String(required=True, indexed=True, internationalizable=True,
                  constraints=[UniqueConstraint(), SizeConstraint(64)])
    description = String(fulltextindexed=True)
    meta = Boolean()
    final = Boolean()



"""Exceptions shared by different ER-Schema modules.

:organization: Logilab
:copyright: 2004-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
:license: General Public License version 2 - http://www.gnu.org/licenses/
"""
__docformat__ = "restructuredtext en"

class SchemaError(Exception):
    """base class for schema exceptions"""

    def __str__(self):
        return unicode(self).encode('utf8')


class UnknownType(SchemaError):
    """using an unknown entity type"""

    msg = 'Unknown type %s'

    def __unicode__(self):
        return self.msg % self.args


class BadSchemaDefinition(SchemaError):
    """error in the schema definition

    instance attributes:
    * filename is the source file where the exception was raised
    * lineno is the line number where the exception was raised
    * line is the actual line in text form
    """

    msg = '%s line %s: %s'

    def __get_filename(self):
        if len(self.args) > 1:
            return self.args[0]
        else:
            return None
    filename = property(__get_filename)

    def __unicode__(self):
        msgs = []
        args_offset = 0
        if self.filename is not None:
            msgs.append(self.filename)
            args_offset += 1
            msgs.append(': ')
        msgs.append(' '.join(self.args[args_offset:]))
        return ''.join(msgs)


class ValidationError(SchemaError):
    """validation error details the reason(s) why the validation failed

    :type entity: EntityType
    :param entity: the entity that could not be validated

    :type explanation: dict
    :param explanation: pairs of (attribute, error)
    """

    def __init__(self, entity, explanation):
        # set args so ValidationError are serializable through pyro
        SchemaError.__init__(self, entity, explanation)
        self.entity = entity
        assert isinstance(explanation, dict), \
            'validation error explanation must be a dict'
        self.errors = explanation

    def __unicode__(self):
        if len(self.errors) == 1:
            attr, error = self.errors.items()[0]
            return u'%s (%s): %s' % (self.entity, attr, error)
        errors = '\n'.join('* %s: %s' % (k, v) for k, v in self.errors.items())
        return u'%s:\n%s' % (self.entity, errors)



from google.appengine.ext import db

class Article(db.Model):
    content = db.TextProperty()
    synopsis = db.StringProperty(default='hello')
    image = db.BlobProperty(required=True)


class Blog(db.Model):
    diem = db.DateProperty(required=True, auto_now_add=True)
    content = db.TextProperty()
    itemtype = db.StringProperty(required=True, choices=('personal', 'business'))
    talks_about = db.ReferenceProperty(Article) 
    cites = db.SelfReferenceProperty() 
    



from yams.buildobjs import String
from yams.reader import context
from data.schema.schema import Affaire, Note

Affaire.permissions = {'read': (),
                       'add': (),
                       'update': (),
                       'delete': ()}

class MyNote(Note):
    text = String()

assert 'Note' in context.defined



"""unit tests for module yams.reader"""

import sys
import os.path as osp
from datetime import datetime, date, time

from logilab.common.testlib import TestCase, unittest_main
from logilab.common.compat import sorted

from yams import BadSchemaDefinition, buildobjs
from yams.schema import Schema
from yams.reader import SchemaLoader
from yams.constraints import StaticVocabularyConstraint, SizeConstraint

from yams import schema
schema.use_py_datetime()

sys.path.insert(0, osp.join(osp.dirname(__file__)))

DATADIR = osp.abspath(osp.join(osp.dirname(__file__), 'data'))

schema = SchemaLoader().load([DATADIR])


class SchemaLoaderTC(TestCase):

    # test helper functions ###################################################

    def test_get_schema_files(self):
        files = [osp.basename(f) for f in SchemaLoader().get_schema_files(DATADIR)]
        self.assertEquals(files[0], '__init__.py')
        self.assertEquals(sorted(files),
                          ['Company.py', 'Dates.py', 'State.py', '__init__.py', 'schema.py'])

    # test load_schema read entity and relation types #######################

    def test_load_schema(self):
        self.assert_(isinstance(schema, Schema))
        self.assertEquals(schema.name, 'NoName')
        self.assertListEquals(sorted(schema.entities()),
                              ['Affaire', 'Boolean', 'Bytes', 'Company', 'Date', 'Datetest', 'Datetime', 'Decimal',
                               'Division', 'EPermission', 'Eetype',  'Employee', 'Float', 'Int', 'Interval',
                               'Note', 'Password', 'Person', 'Societe', 'State', 'String',
                               'Subcompany', 'Subdivision', 'Time', 'pkginfo'])
        self.assertListEquals(sorted(schema.relations()),
                              ['ad1', 'ad2', 'ad3', 'adel', 'ass', 'author', 'author_email',
                               'concerne', 'copyright', 'cp',
                               'd1', 'd2', 'date', 'datenaiss', 'debian_handler', 'description', 'division_of', 'dt1', 'dt2',
                               'eid', 'evaluee', 'fax', 'final',
                               'initial_state', 'inline_rel',
                               'license', 'long_desc',
                               'mailinglist', 'meta', 'modname',
                               'name', 'next_state', 'nom', 'obj_wildcard',
                               'para', 'prenom', 'promo',
                               'ref', 'require_permission', 'rncs',
                               'salary', 'sexe', 'short_desc', 'state_of', 'subcompany_of',
                               'subdivision_of', 'subj_wildcard', 'sujet', 'sym_rel',
                               't1', 't2', 'tel', 'test', 'titre', 'travaille', 'type',
                               'version',
                               'ville', 'web', 'works_for'])

    def test_eschema(self):
        eschema = schema.eschema('Societe')
        self.assertEquals(eschema.description, '')
        self.assertEquals(eschema.is_final(), False)
        self.assertListEquals(sorted(eschema.subject_relations()),
                              ['ad1', 'ad2', 'ad3', 'cp', 'evaluee',
                               'fax', 'nom', 'rncs', 'subj_wildcard', 'tel', 'ville',
                               'web'])
        self.assertListEquals(sorted(eschema.object_relations()),
                          ['concerne', 'obj_wildcard', 'travaille'])

        eschema = schema.eschema('Eetype')
        self.assertEquals(eschema.description, 'define an entity type, used to build the application schema')
        self.assertEquals(eschema.is_final(), False)
        self.assertListEquals(sorted(str(r) for r in eschema.subject_relations()),
                              ['description', 'final', 'initial_state', 'meta',
                               'name', 'subj_wildcard'])
        self.assertListEquals(sorted(str(r) for r in eschema.object_relations()),
                              ['obj_wildcard', 'state_of'])

        eschema = schema.eschema('Boolean')
        self.assertEquals(eschema.description, '')
        self.assertEquals(eschema.is_final(), True)
        self.assertListEquals(sorted(eschema.subject_relations()),
                              [])
        self.assertListEquals(sorted(eschema.object_relations()),
                              ['final', 'meta', 'test'])

    # test base entity type's subject relation properties #####################

    def test_indexed(self):
        eschema = schema.eschema('Person')
        self.assert_(not eschema.rproperty('nom', 'indexed'))
        eschema = schema.eschema('State')
        self.assert_(eschema.rproperty('name', 'indexed'))

    def test_uid(self):
        eschema = schema.eschema('State')
        self.assert_(eschema.rproperty('eid', 'uid'))
        self.assert_(not eschema.rproperty('name', 'uid'))

    def test_fulltextindexed(self):
        eschema = schema.eschema('Person')
        self.assert_(not eschema.rproperty('tel', 'fulltextindexed'))
        self.assert_(eschema.rproperty('nom', 'fulltextindexed'))
        self.assert_(eschema.rproperty('prenom', 'fulltextindexed'))
        self.assert_(not eschema.rproperty('sexe', 'fulltextindexed'))
        indexable = sorted(eschema.indexable_attributes())
        self.assertEquals(['nom', 'prenom', 'titre'], indexable)
        self.assertEquals(schema.rschema('works_for').fulltext_container, None)
        self.assertEquals(schema.rschema('require_permission').fulltext_container,
                          'subject')
        eschema = schema.eschema('Company')
        indexable = sorted(eschema.indexable_attributes())
        self.assertEquals([], indexable)
        indexable = sorted(eschema.fulltext_relations())
        self.assertEquals([('require_permission', 'subject')], indexable)
        containers = sorted(eschema.fulltext_containers())
        self.assertEquals([], containers)
        eschema = schema.eschema('EPermission')
        indexable = sorted(eschema.indexable_attributes())
        self.assertEquals(['name'], indexable)
        indexable = sorted(eschema.fulltext_relations())
        self.assertEquals([], indexable)
        containers = sorted(eschema.fulltext_containers())
        self.assertEquals([('require_permission', 'subject')], containers)

    def test_internationalizable(self):
        eschema = schema.eschema('Eetype')
        self.assert_(eschema.rproperty('name', 'internationalizable'))
        eschema = schema.eschema('State')
        self.assert_(eschema.rproperty('name', 'internationalizable'))
        eschema = schema.eschema('Societe')
        self.assert_(not eschema.rproperty('ad1', 'internationalizable'))

    # test advanced entity type's subject relation properties #################

    def test_vocabulary(self):
        eschema = schema.eschema('pkginfo')
        self.assertEquals(eschema.vocabulary('license'), ('GPL', 'ZPL'))
        self.assertEquals(eschema.vocabulary('debian_handler'), ('machin', 'bidule'))

    def test_default(self):
        eschema = schema.eschema('pkginfo')
        self.assertEquals(eschema.default('version'), '0.1')
        self.assertEquals(eschema.default('license'), None)

    # test relation type properties ###########################################

    def test_rschema(self):
        rschema = schema.rschema('evaluee')
        self.assertEquals(rschema.symetric, False)
        self.assertEquals(rschema.description, '')
        self.assertEquals(rschema.is_final(), False)
        self.assertListEquals(sorted(rschema.subjects()), ['Person', 'Societe'])
        self.assertListEquals(sorted(rschema.objects()), ['Note'])

        rschema = schema.rschema('sym_rel')
        self.assertEquals(rschema.symetric, True)
        self.assertEquals(rschema.description, '')
        self.assertEquals(rschema.is_final(), False)
        self.assertListEquals(sorted(rschema.subjects()), ['Affaire', 'Person'])
        self.assertListEquals(sorted(rschema.objects()), ['Affaire', 'Person'])

        rschema = schema.rschema('initial_state')
        self.assertEquals(rschema.symetric, False)
        self.assertEquals(rschema.description, 'indicate which state should be used by default when an entity using states is created')
        self.assertEquals(rschema.is_final(), False)
        self.assertListEquals(sorted(rschema.subjects()), ['Eetype'])
        self.assertListEquals(sorted(rschema.objects()), ['State'])

        rschema = schema.rschema('name')
        self.assertEquals(rschema.symetric, False)
        self.assertEquals(rschema.description, '')
        self.assertEquals(rschema.is_final(), True)
        self.assertListEquals(sorted(rschema.subjects()), ['Company', 'Division', 'EPermission', 'Eetype', 'State', 'Subcompany', 'Subdivision'])
        self.assertListEquals(sorted(rschema.objects()), ['String'])

    def test_cardinality(self):
        rschema = schema.rschema('evaluee')
        self.assertEquals(rschema.rproperty('Person', 'Note', 'cardinality'), '**')
        rschema = schema.rschema('inline_rel')
        self.assertEquals(rschema.rproperty('Affaire', 'Person', 'cardinality'), '?*')
        rschema = schema.rschema('initial_state')
        self.assertEquals(rschema.rproperty('Eetype', 'State', 'cardinality'), '?*')
        rschema = schema.rschema('state_of')
        self.assertEquals(rschema.rproperty('State', 'Eetype', 'cardinality'), '+*')
        rschema = schema.rschema('name')
        self.assertEquals(rschema.rproperty('State', 'String', 'cardinality'), '11')
        rschema = schema.rschema('description')
        self.assertEquals(rschema.rproperty('State', 'String', 'cardinality'), '?1')

    def test_constraints(self):
        eschema = schema.eschema('Person')
        self.assertEquals(len(eschema.constraints('nom')), 1)
        self.assertEquals(len(eschema.constraints('promo')), 2)
        self.assertEquals(len(eschema.constraints('tel')), 0)
        self.assertRaises(AssertionError, eschema.constraints, 'travaille')
        self.assertRaises(KeyError, eschema.constraints, 'inline_rel')
        eschema = schema.eschema('State')
        self.assertEquals(len(eschema.constraints('name')), 1)
        self.assertEquals(len(eschema.constraints('description')), 0)
        self.assertRaises(AssertionError, eschema.constraints, 'next_state')
        self.assertRaises(KeyError, eschema.constraints, 'initial_state')
        eschema = schema.eschema('Eetype')
        self.assertEquals(len(eschema.constraints('name')), 2)

    def test_inlined(self):
        rschema = schema.rschema('evaluee')
        self.assertEquals(rschema.inlined, False)
        rschema = schema.rschema('state_of')
        self.assertEquals(rschema.inlined, False)
        rschema = schema.rschema('inline_rel')
        self.assertEquals(rschema.inlined, True)
        rschema = schema.rschema('initial_state')
        self.assertEquals(rschema.inlined, True)

    def test_relation_permissions(self):
        rschema = schema.rschema('state_of')
        self.assertEquals(rschema._groups,
                          {'read': ('managers', 'users', 'guests'),
                           'delete': ('managers',),
                           'add': ('managers',)})

        rschema = schema.rschema('next_state')
        self.assertEquals(rschema._groups,
                          {'read':   ('managers', 'users', 'guests',),
                           'add':    ('managers',),
                           'delete': ('managers',)})

        rschema = schema.rschema('initial_state')
        self.assertEquals(rschema._groups,
                          {'read':   ('managers', 'users', 'guests',),
                           'add':    ('managers', 'users',),
                           'delete': ('managers', 'users',)})

        rschema = schema.rschema('evaluee')
        self.assertEquals(rschema._groups,
                          {'read':   ('managers', 'users', 'guests',),
                           'add':    ('managers', 'users',),
                           'delete': ('managers', 'users',)})

        rschema = schema.rschema('nom')
        self.assertEquals(rschema._groups, {'read': ('managers', 'users', 'guests'),
                                            'add': ('managers', 'users', 'guests'),
                                            'delete': ('managers', 'users', 'guests')})

        rschema = schema.rschema('require_permission')
        self.assertEquals(rschema._groups, {'read': ('managers', 'users', 'guests'),
                                            'add': ('managers', ),
                                            'delete': ('managers',)})

    def test_entity_permissions(self):
        eschema = schema.eschema('State')
        self.assertEquals(eschema._groups,
                          {'read':   ('managers', 'users', 'guests',),
                           'add':    ('managers', 'users',),
                           'delete': ('managers', 'owners',),
                           'update': ('managers', 'owners',)})

        eschema = schema.eschema('Eetype')
        self.assertEquals(eschema._groups,
                          {'read':   ('managers', 'users', 'guests',),
                           'add':    ('managers',),
                           'delete': ('managers',),
                           'update': ('managers', 'owners',)})

        eschema = schema.eschema('Person')
        self.assertEquals(eschema._groups,
                          {'read':   ('managers', 'users', 'guests',),
                           'add':    ('managers', 'users',),
                           'delete': ('managers', 'owners',),
                           'update': ('managers', 'owners',)})

##     def test_nonregr_using_tuple_as_relation_target(self):
##         rschema = schema.rschema('see_also')
##         self.assertEquals(rschema.symetric, False)
##         self.assertEquals(rschema.description, '')
##         self.assertEquals(rschema.is_final(), False)
##         self.assertListEquals(sorted(rschema.subjects()), ['Employee'])
##         self.assertListEquals(sorted(rschema.objects()), ['Company', 'Division'])
##


from yams import buildobjs as B

class BasePerson(B.EntityType):
    firstname = B.String(vocabulary=('logilab', 'caesium'), maxsize=10)
    lastname = B.String(constraints=[StaticVocabularyConstraint(['logilab', 'caesium'])])

class Person(BasePerson):
    email = B.String()

class Employee(Person):
    company = B.String(vocabulary=('logilab', 'caesium'))


class Student(Person):
    __specializes_schema__ = True
    college = B.String()

class X(Student):
    pass

class Foo(B.EntityType):
    i = B.Int(required=True)
    f = B.Float()
    d = B.Datetime()


class PySchemaTC(TestCase):

    def test_python_inheritance(self):
        bp = BasePerson()
        p = Person()
        e = Employee()
        self.assertEquals([r.name for r in bp.__relations__], ['firstname', 'lastname'])
        self.assertEquals([r.name for r in p.__relations__], ['firstname', 'lastname', 'email'])
        self.assertEquals([r.name for r in e.__relations__], ['firstname', 'lastname', 'email', 'company'])

    def test_schema_extension(self):
        s = Student()
        self.assertEquals([r.name for r in s.__relations__], ['firstname', 'lastname', 'email', 'college'])
        self.assertEquals(s.specialized_type, 'Person')
        x = X()
        self.assertEquals(x.specialized_type, None)

    def test_relationtype(self):
        foo = Foo()
        self.assertEquals([r.etype for r in foo.__relations__],
                          ['Int', 'Float', 'Datetime'])
        self.assertEquals(foo.__relations__[0].cardinality, '11')
        self.assertEquals(foo.__relations__[1].cardinality, '?1')

    def test_maxsize(self):
        bp = BasePerson()
        def maxsize(e):
            for e in e.constraints:
                if isinstance(e, SizeConstraint):
                    return e.max
        self.assertEquals(maxsize(bp.__relations__[0]), 7)
        # self.assertEquals(maxsize(bp.__relations__[1]), 7)
        emp = Employee()
        self.assertEquals(maxsize(emp.__relations__[3]), 7)

    def test_date_defaults(self):
        _today = date.today()
        _now = datetime.now()
        datetest = schema.eschema('Datetest')
        dt1 = datetest.default('dt1')
        dt2 = datetest.default('dt2')
        d1 = datetest.default('d1')
        d2 = datetest.default('d2')
        t1 = datetest.default('t1')
        t2 = datetest.default('t2')
        # datetimes
        self.assertIsInstance(dt1, datetime)
        # there's no easy way to test NOW (except monkey patching now() itself)
        delta = dt1 - _now
        self.failUnless(abs(delta.seconds) < 5)
        self.assertEquals(date(dt2.year, dt2.month, dt2.day), _today)
        self.assertIsInstance(dt2, datetime)
        # dates
        self.assertEquals(d1, _today)
        self.assertIsInstance(d1, date)
        self.assertEquals(d2, datetime(2007, 12, 11, 0, 0))
        self.assertIsInstance(d2, datetime)
        # times
        self.assertEquals(t1, time(8, 40))
        self.assertIsInstance(t1, time)
        self.assertEquals(t2, time(9, 45))
        self.assertIsInstance(t2, time)


class SchemaLoaderTC2(TestCase):
    def test_broken_schema1(self):
        SchemaLoader.main_schema_directory = 'brokenschema1'
        ex = self.assertRaises(BadSchemaDefinition,
                               SchemaLoader().load, [DATADIR], 'Test')
        self.assertEquals(str(ex), "conflicting values False/True for property inlined of relation type 'rel'")

    def test_broken_schema2(self):
        SchemaLoader.main_schema_directory = 'brokenschema2'
        ex = self.assertRaises(BadSchemaDefinition,
                               SchemaLoader().load, [DATADIR], 'Test')
        self.assertEquals(str(ex), "conflicting values True/False for property inlined of relation type 'rel'")

    def test_broken_schema3(self):
        SchemaLoader.main_schema_directory = 'brokenschema3'
        ex = self.assertRaises(BadSchemaDefinition,
                               SchemaLoader().load, [DATADIR], 'Test')
        self.assertEquals(str(ex), "conflicting values True/False for property inlined of relation type 'rel'")

    def test_broken_schema4(self):
        schema = Schema('toto')
        schema.add_entity_type(buildobjs.EntityType(name='Entity'))
        schema.add_entity_type(buildobjs.EntityType(name='Int'))
        schema.add_relation_type(buildobjs.RelationType(name='toto'))
        ex = self.assertRaises(BadSchemaDefinition,
                               schema.add_relation_def,
                               buildobjs.RelationDefinition(name='toto', subject='Entity', object='Int',
                                                     constraints=[SizeConstraint(40)]))
        self.assertEquals(str(ex), "size constraint doesn't apply to Int entity type")
        
    def test_broken_schema5(self):
        schema = Schema('toto')
        schema.add_entity_type(buildobjs.EntityType(name='Entity'))
        schema.add_entity_type(buildobjs.EntityType(name='String'))
        schema.add_relation_type(buildobjs.RelationType(name='toto'))
        ex = self.assertRaises(BadSchemaDefinition,
                               schema.add_relation_def,
                               buildobjs.RelationDefinition(name='toto', subject='Entity', object='String',
                                                            constraints=[StaticVocabularyConstraint(['ab', 'abc']),
                                                                         SizeConstraint(2)]))
        self.assertEquals(str(ex), "size constraint set to 2 but vocabulary contains string of greater size")

    def test_schema(self):
        SchemaLoader.main_schema_directory = 'schema2'
        schema = SchemaLoader().load([DATADIR], 'Test')
        rel = schema['rel']
        self.assertEquals(rel.rproperty('Anentity', 'Anentity', 'composite'),
                          'subject')
        self.assertEquals(rel.rproperty('Anotherentity', 'Anentity', 'composite'),
                          'subject')
        self.assertEquals(rel.rproperty('Anentity', 'Anentity', 'cardinality'),
                          '1*')
        self.assertEquals(rel.rproperty('Anotherentity', 'Anentity', 'cardinality'),
                          '1*')
        self.assertEquals(rel.symetric, True)
        self.assertEquals(rel.inlined, True)

    def test_imports(self):
        SchemaLoader.main_schema_directory = 'schema'
        schema = SchemaLoader().load([DATADIR, DATADIR+'2'], 'Test')
        self.assertEquals(schema['Affaire']._groups, {'read': (),
                                                      'add': (),
                                                      'update': (),
                                                      'delete': ()})
        self.assertEquals([str(r) for r,at in schema['MyNote'].attribute_definitions()],
                          ['date', 'type', 'para', 'text'])

if __name__ == '__main__':
    unittest_main()





from yams.buildobjs import EntityType, SubjectRelation
# conflicting RelationType properties

class Anentity(EntityType):
    rel = SubjectRelation('Anentity', inlined=True)

class Anotherentity(EntityType):
    rel = SubjectRelation('Anentity', inlined=False)    



from yams.buildobjs import (EntityType, SubjectRelation, ObjectRelation,
                            String, Int, Float, Date, Boolean)

class Affaire(EntityType):
    sujet = String(maxsize=128)
    ref = String(maxsize=12)

    concerne = SubjectRelation('Societe')
    obj_wildcard = SubjectRelation('*')
    sym_rel = SubjectRelation('Person', symetric=True)
    inline_rel = SubjectRelation('Person', inlined=True, cardinality='?*')
    subj_wildcard = ObjectRelation('*')


class Person(EntityType):
    nom    = String(maxsize=64, fulltextindexed=True, required=True)
    prenom = String(maxsize=64, fulltextindexed=True)
    sexe   = String(maxsize=1, default='M')
    promo  = String(vocabulary=('bon','pasbon'))
    titre  = String(maxsize=128, fulltextindexed=True)
    adel   = String(maxsize=128)
    ass    = String(maxsize=128)
    web    = String(maxsize=128)
    tel    = Int()
    fax    = Int()
    datenaiss = Date()
    test   = Boolean()
    salary = Float()

    travaille = SubjectRelation('Societe')
    evaluee = SubjectRelation('Note')
    concerne = SubjectRelation('Affaire')


class Societe(EntityType):
    meta = False # test bw compat
    nom  = String(maxsize=64, fulltextindexed=True)
    web = String(maxsize=128)
    tel  = Int()
    fax  = Int()
    rncs = String(maxsize=32)
    ad1  = String(maxsize=128)
    ad2  = String(maxsize=128)
    ad3  = String(maxsize=128)
    cp   = String(maxsize=12)
    ville = String(maxsize=32)

    evaluee = SubjectRelation('Note')


class Note(EntityType):
    date = String(maxsize=10)
    type = String(maxsize=1)
    para = String(maxsize=512)


class pkginfo(EntityType):
    modname = String(maxsize=30, required=True)
    version = String(maxsize=10, required=True, default='0.1')
    copyright = String(required=True)
    license = String(vocabulary=('GPL', 'ZPL'))
    short_desc = String(maxsize=80, required=True)
    long_desc = String(required=True, fulltextindexed=True)
    author = String(maxsize=100, required=True)
    author_email = String(maxsize=100, required=True)
    mailinglist = String(maxsize=100)
    debian_handler = String(vocabulary=('machin', 'bidule'))




"""Yams packaging information.

:organization: Logilab
:copyright: 2003-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
:license: General Public License version 2 - http://www.gnu.org/licenses
"""
__docformat__ = "restructuredtext en"

# pylint: disable-msg=W0622

# package name
modname = 'yams'

# release version
numversion = (0, 24, 0)
version = '.'.join(str(num) for num in numversion)

# license and copyright
license = 'GPL'
copyright = 'Copyright (c) 2003-2009 LOGILAB S.A. (Paris, FRANCE)'

# short and long description
short_desc = "entity / relation schema"
long_desc = """Yet Another Magic Schema !
A simple/generic but powerful entities / relations schema, suitable
to represent RDF like data. The schema is readable/writable from/to
various formats.
"""

# author name and email
author = "Logilab"
author_email = "devel@logilab.fr"

# home page
web = "http://www.logilab.org/project/%s" % modname

# mailing list
mailinglist = 'mailto://python-projects@lists.logilab.org'

# download place
ftp = "ftp://ftp.logilab.org/pub/%s" % modname

# is there some directories to include with the source installation
include_dirs = []

# executable

scripts = ['bin/yams-check', 'bin/yams-view']

pyversions = ['2.4', '2.5']



from yams.buildobjs import EntityType, RelationType, SubjectRelation
# conflicting RelationType properties

class Anentity(EntityType):
    rel = SubjectRelation('Anentity', inlined=True)

class rel(RelationType):
    inlined = False
